from typing import Any, Dict

from torch import nn

from oml.losses.arcface import ArcFaceLoss, ArcFaceLossWithMLP
from oml.losses.surrogate_precision import SurrogatePrecision
from oml.losses.triplet import TripletLoss, TripletLossPlain, TripletLossWithMiner
from oml.registry.miners import get_miner_by_cfg
from oml.utils.misc import TCfg, dictconfig_to_dict, remove_unused_kwargs

LOSSES_REGISTRY = {
    "triplet": TripletLoss,
    "triplet_plain": TripletLossPlain,
    "triplet_with_miner": TripletLossWithMiner,
    "surrogate_precision": SurrogatePrecision,
    "arcface": ArcFaceLoss,
    "mlp_arcface": ArcFaceLossWithMLP,
}


def get_criterion(name: str, **kwargs: Dict[str, Any]) -> nn.Module:
    constructor = LOSSES_REGISTRY[name]
    if "miner" in kwargs:
        miner = get_miner_by_cfg(kwargs.pop("miner"))
        kwargs = remove_unused_kwargs(kwargs, constructor)
        return constructor(miner=miner, **kwargs)
    else:
        kwargs = remove_unused_kwargs(kwargs, constructor)
        return constructor(**kwargs)


def get_criterion_by_cfg(cfg: TCfg, **kwargs_runtime: Dict[str, Any]) -> nn.Module:
    cfg = dictconfig_to_dict(cfg)
    cfg.setdefault("args", {})
    cfg["args"].update(**kwargs_runtime)
    return get_criterion(name=cfg["name"], **cfg["args"])


__all__ = ["LOSSES_REGISTRY", "get_criterion", "get_criterion_by_cfg"]



import os
import tempfile
from pathlib import Path
from sys import platform
from typing import Any, Dict, Optional, Sequence, Tuple, Union

from omegaconf import DictConfig


def get_cache_folder() -> Path:
    if platform == "linux" or platform == "linux2":
        return Path(os.environ.get("XDG_CACHE_HOME", Path.home() / ".cache")) / "oml"

    elif platform == "darwin":  # mac os
        return Path.home() / "Library" / "Caches" / "oml"

    elif platform.startswith("win"):
        return Path.home() / ".cache" / "oml"

    else:
        raise ValueError(f"Unexpected platform {platform}.")


OML_PATH = Path(__file__).parent
PROJECT_ROOT = OML_PATH.parent
CACHE_PATH = get_cache_folder()
TMP_PATH = Path(tempfile.gettempdir())

DOTENV_PATH = PROJECT_ROOT / ".env"
CONFIGS_PATH = PROJECT_ROOT / "oml" / "configs"

CKPT_SAVE_ROOT = CACHE_PATH / "torch" / "checkpoints"

LOG_IMAGE_FOLDER = "image_logs"
LOG_TOPK_ROWS_PER_METRIC = 5
LOG_TOPK_IMAGES_PER_ROW = 5
N_GT_SHOW_EMBEDDING_METRICS = 2

STORAGE_URL = "https://oml.daloroserver.com"
STORAGE_CKPTS = STORAGE_URL + "/download/checkpoints"

MOCK_DATASET_PATH = CACHE_PATH / "mock_dataset"
MOCK_DATASET_URL_GDRIVE = "https://drive.google.com/drive/folders/1plPnwyIkzg51-mLUXWTjREHgc1kgGrF4?usp=sharing"
MOCK_DATASET_MD5 = "a23478efd4746d18f937fa6c5758c0ed"

REQUESTS_TIMEOUT = 120.0

TColor = Tuple[int, int, int]
RED = (255, 0, 0)
GREEN = (0, 255, 0)
BLUE = (0, 0, 255)
GRAY = (120, 120, 120)
BLACK = (0, 0, 0)
PAD_COLOR = (255, 255, 255)

BS_KNN = 5_000

TCfg = Union[Dict[str, Any], DictConfig]

# ImageNet Params
TNormParam = Tuple[float, float, float]
MEAN: TNormParam = (0.485, 0.456, 0.406)
STD: TNormParam = (0.229, 0.224, 0.225)

MEAN_CLIP = (0.48145466, 0.4578275, 0.40821073)
STD_CLIP = (0.26862954, 0.26130258, 0.27577711)

TBBox = Tuple[int, int, int, int]
TBBoxes = Sequence[Optional[TBBox]]

CROP_KEY = "crop"  # the format is [x1, y1, x2, y2]

# Required dataset format:
LABELS_COLUMN = "label"
SPLIT_COLUMN = "split"
IS_QUERY_COLUMN = "is_query"
IS_GALLERY_COLUMN = "is_gallery"
CATEGORIES_COLUMN = "category"
SEQUENCE_COLUMN = "sequence"
# text specific columns:
TEXTS_COLUMN = "text"
# image specific columns:
PATHS_COLUMN = "path"
X1_COLUMN = "x_1"
X2_COLUMN = "x_2"
Y1_COLUMN = "y_1"
Y2_COLUMN = "y_2"

OBLIGATORY_COLUMNS = [LABELS_COLUMN, SPLIT_COLUMN, IS_QUERY_COLUMN, IS_GALLERY_COLUMN]
BBOXES_COLUMNS = [X1_COLUMN, X2_COLUMN, Y1_COLUMN, Y2_COLUMN]

# Keys for interactions among our classes (datasets, metrics and so on)
OVERALL_CATEGORIES_KEY = "OVERALL"
INPUT_TENSORS_KEY = "input_tensors"
LABELS_KEY = "labels"
EMBEDDINGS_KEY = "embeddings"
INDEX_KEY = "idx"

INPUT_TENSORS_KEY_1 = "input_tensors_1"
INPUT_TENSORS_KEY_2 = "input_tensors_2"

IMAGE_EXTENSIONS = ["jpg", "jpeg", "JPG", "JPEG", "png"]

# hydra provides ability to set its behaviour for convinient backwards compatibility
# no matter what the current version is
HYDRA_BEHAVIOUR = "1.1"



import hashlib
from pathlib import Path
from pprint import pprint
from typing import Any, Dict, Tuple

import pytorch_lightning as pl
import torch
from omegaconf import DictConfig
from torch import device as tdevice
from torch.utils.data import DataLoader

from oml.const import EMBEDDINGS_KEY, TCfg
from oml.datasets.base import ImageLabeledDataset, ImageQueryGalleryLabeledDataset
from oml.datasets.images import get_retrieval_images_datasets
from oml.inference import inference, inference_cached
from oml.interfaces.datasets import ILabeledDataset, IQueryGalleryLabeledDataset
from oml.interfaces.models import IPairwiseModel
from oml.lightning.callbacks.metric import MetricValCallback
from oml.lightning.modules.pairwise_postprocessing import (
    PairwiseModule,
    PairwiseModuleDDP,
)
from oml.lightning.pipelines.parser import (
    check_is_config_for_ddp,
    parse_ckpt_callback_from_config,
    parse_engine_params_from_config,
    parse_logger_from_config,
    parse_sampler_from_config,
    parse_scheduler_from_config,
)
from oml.metrics.embeddings import EmbeddingMetrics
from oml.miners.pairs import PairsMiner
from oml.registry.models import get_extractor_by_cfg
from oml.registry.optimizers import get_optimizer_by_cfg
from oml.registry.postprocessors import get_postprocessor_by_cfg
from oml.registry.transforms import get_transforms_by_cfg
from oml.retrieval.postprocessors.pairwise import PairwiseReranker
from oml.utils.misc import dictconfig_to_dict, flatten_dict, set_global_seed


def get_hash_of_extraction_stage_cfg(cfg: TCfg) -> str:
    def dict2str(dictionary: Dict[str, Any]) -> str:
        flatten_items = flatten_dict(dictionary).items()
        sorted(flatten_items, key=lambda x: x[0])
        return str(flatten_items)

    cfg_extraction_str = (
        dict2str(cfg["extractor"])
        + dict2str(cfg["transforms_extraction"])
        + str(cfg["dataframe_name"])
        + str(cfg.get("precision", 32))
    )

    md5sum = hashlib.md5(cfg_extraction_str.encode("utf-8")).hexdigest()
    return md5sum


def get_loaders_with_embeddings(
    cfg: TCfg,
) -> Tuple[DataLoader, DataLoader, ILabeledDataset, IQueryGalleryLabeledDataset]:
    device = tdevice("cuda:0") if parse_engine_params_from_config(cfg)["accelerator"] == "gpu" else tdevice("cpu")
    extractor = get_extractor_by_cfg(cfg["extractor"]).to(device)

    transforms_extraction = get_transforms_by_cfg(cfg["transforms_extraction"])

    train_extraction, val_extraction = get_retrieval_images_datasets(
        dataset_root=Path(cfg["dataset_root"]),
        dataframe_name=cfg["dataframe_name"],
        transforms_train=transforms_extraction,
        transforms_val=transforms_extraction,
    )

    args = {
        "model": extractor,
        "num_workers": cfg["num_workers"],
        "batch_size": cfg["batch_size_inference"],
        "use_fp16": int(cfg.get("precision", 32)) == 16,
    }

    if cfg["embeddings_cache_dir"] is not None:
        hash_ = get_hash_of_extraction_stage_cfg(cfg)[:5]
        dir_ = Path(cfg["embeddings_cache_dir"])
        emb_train = inference_cached(dataset=train_extraction, cache_path=str(dir_ / f"emb_train_{hash_}.pkl"), **args)
        emb_val = inference_cached(dataset=val_extraction, cache_path=str(dir_ / f"emb_val_{hash_}.pkl"), **args)
    else:
        emb_train = inference(dataset=train_extraction, **args)
        emb_val = inference(dataset=val_extraction, **args)

    train_dataset = ImageLabeledDataset(
        dataset_root=cfg["dataset_root"],
        cache_size=cfg.get("cache_size", 0),
        df=train_extraction.df,
        transform=get_transforms_by_cfg(cfg["transforms_train"]),
        extra_data={EMBEDDINGS_KEY: emb_train},
    )

    valid_dataset = ImageQueryGalleryLabeledDataset(
        dataset_root=cfg["dataset_root"],
        cache_size=cfg.get("cache_size", 0),
        df=val_extraction.df,
        transform=transforms_extraction,
        extra_data={EMBEDDINGS_KEY: emb_val},
    )

    sampler = parse_sampler_from_config(cfg, dataset=train_dataset)
    assert sampler is not None, "We will be training on pairs, so, having sampler is obligatory."

    loader_train = DataLoader(batch_sampler=sampler, dataset=train_dataset, num_workers=cfg["num_workers"])

    loader_val = DataLoader(
        dataset=valid_dataset, batch_size=cfg["batch_size_inference"], num_workers=cfg["num_workers"], shuffle=False
    )

    return loader_train, loader_val, train_dataset, valid_dataset


def postprocessor_training_pipeline(cfg: DictConfig) -> None:
    """
    This pipeline allows you to train and validate a pairwise postprocessor
    which fixes mistakes of a feature extractor in retrieval setup.

    The config can be specified as a dictionary or with hydra: https://hydra.cc/.
    For more details look at ``pipelines/postprocessing/pairwise_postprocessing/README.md``

    """
    set_global_seed(cfg["seed"])

    cfg = dictconfig_to_dict(cfg)
    pprint(cfg)

    logger = parse_logger_from_config(cfg)
    logger.log_pipeline_info(cfg)

    trainer_engine_params = parse_engine_params_from_config(cfg)
    is_ddp = check_is_config_for_ddp(trainer_engine_params)

    loader_train, loader_val, dataset_train, dataset_val = get_loaders_with_embeddings(cfg)

    postprocessor = None if not cfg.get("postprocessor", None) else get_postprocessor_by_cfg(cfg["postprocessor"])
    assert isinstance(postprocessor, PairwiseReranker), f"We only support {PairwiseReranker.__name__} at the moment."
    assert isinstance(postprocessor.model, IPairwiseModel), f"You model must be a child of {IPairwiseModel.__name__}"

    criterion = torch.nn.BCEWithLogitsLoss()
    pairs_miner = PairsMiner(hard_mining=cfg["hard_pairs_mining"])
    optimizer = get_optimizer_by_cfg(cfg["optimizer"], **{"params": postprocessor.model.parameters()})

    module_kwargs = {}
    module_kwargs.update(parse_scheduler_from_config(cfg, optimizer=optimizer))
    if is_ddp:
        module_kwargs.update({"loaders_train": loader_train, "loaders_val": loader_val})
        module_constructor = PairwiseModuleDDP
    else:
        module_constructor = PairwiseModule  # type: ignore

    pl_module = module_constructor(
        pairwise_model=postprocessor.model,
        pairs_miner=pairs_miner,
        criterion=criterion,
        optimizer=optimizer,
        input_tensors_key=dataset_train.input_tensors_key,
        labels_key=dataset_train.labels_key,
        embeddings_key=EMBEDDINGS_KEY,
        freeze_n_epochs=cfg.get("freeze_n_epochs", 0),
        **module_kwargs,
    )

    metric = EmbeddingMetrics(dataset=dataset_val, postprocessor=postprocessor, **cfg.get("metric_args", {}))
    metrics_clb = MetricValCallback(metric=metric, log_images=cfg.get("log_images", True))

    trainer = pl.Trainer(
        max_epochs=cfg["max_epochs"],
        num_sanity_val_steps=0,
        check_val_every_n_epoch=cfg["valid_period"],
        default_root_dir=str(Path.cwd()),
        enable_checkpointing=True,
        enable_progress_bar=True,
        enable_model_summary=True,
        precision=cfg.get("precision", 32),
        logger=logger,
        callbacks=[metrics_clb, parse_ckpt_callback_from_config(cfg)],
        **trainer_engine_params,
    )

    if is_ddp:
        trainer.fit(model=pl_module)
    else:
        trainer.fit(model=pl_module, train_dataloaders=loader_train, val_dataloaders=loader_val)


__all__ = ["postprocessor_training_pipeline"]



from typing import Any, Dict

import torch.optim as opt

from oml.utils.misc import TCfg, dictconfig_to_dict

OPTIMIZERS_REGISTRY = {
    "adadelta": opt.Adadelta,
    "adagrad": opt.Adagrad,
    "adam": opt.Adam,
    "adamw": opt.AdamW,
    "adamax": opt.Adamax,
    "asgd": opt.ASGD,
    "lbfgs": opt.LBFGS,
    "rmsprop": opt.RMSprop,
    "rprop": opt.Rprop,
    "sgd": opt.SGD,
}


def get_optimizer(name: str, **kwargs: Dict[str, Any]) -> opt.Optimizer:
    return OPTIMIZERS_REGISTRY[name](**kwargs)


def get_optimizer_by_cfg(cfg: TCfg, **kwargs_runtime: Dict[str, Any]) -> opt.Optimizer:
    cfg = dictconfig_to_dict(cfg)
    cfg["args"].update(kwargs_runtime)
    return get_optimizer(name=cfg["name"], **cfg["args"])


__all__ = ["OPTIMIZERS_REGISTRY", "get_optimizer", "get_optimizer_by_cfg"]



import torch
from torch import Tensor, sigmoid


def get_reduced(tensor: Tensor, reduction: str) -> Tensor:
    if reduction == "mean":
        return tensor.mean()

    elif reduction == "sum":
        return tensor.sum()

    elif reduction == "none":
        return tensor

    else:
        raise ValueError(f"Unexpected type of reduction: {reduction}")


def surrogate_precision(
    distances: Tensor, mask_gt: Tensor, k: int, t1: float, t2: float, reduction: str = "mean"
) -> Tensor:
    distances_diff = distances - distances.unsqueeze(0).permute(2, 1, 0)
    rank = sigmoid(distances_diff / t2).sum(dim=0)
    precision = (sigmoid((k - rank) / t1) * mask_gt).sum(dim=1) / torch.clip(mask_gt.sum(dim=1), max=k)
    return get_reduced(precision, reduction=reduction)



import inspect
import os
import random
from contextlib import contextmanager
from typing import Any, Dict, Generator, Iterable, List, Sequence, Tuple, Union

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import torch
from omegaconf import DictConfig, OmegaConf
from torch import Tensor

from oml.const import BLACK, TCfg, TColor


def find_value_ids(it: Iterable[Any], value: Any) -> List[int]:
    """
    Args:
        it: List of any
        value: Query element

    Returns:
        Indices of the all elements equal to x0
    """
    if isinstance(it, np.ndarray):
        inds = list(np.where(it == value)[0])
    else:  # could be very slow
        inds = [i for i, el in enumerate(it) if el == value]
    return inds


def set_global_seed(seed: int) -> None:
    random.seed(seed)
    np.random.seed(seed)

    torch.manual_seed(seed)
    if torch.cuda.is_available():
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)
        torch.backends.cudnn.benchmark = False
        torch.backends.cudnn.deterministic = True

    os.environ["PL_SEED_WORKERS"] = str(1)

    try:
        import torch_xla.core.xla_model as xm
    except ImportError:
        pass
    else:
        xm.set_rng_state(seed)


def one_hot(i: int, dim: int) -> torch.Tensor:
    vector = torch.zeros(dim)
    vector[i] = 1
    return vector


def flatten_dict(
    d: Dict[str, Any], parent_key: str = "", sep: str = "/", ignored_keys: Iterable[str] = ()
) -> Dict[str, Any]:
    items = []  # type: ignore
    for k, v in d.items():
        if k in ignored_keys:
            continue
        new_key = str(parent_key) + sep + str(k) if parent_key else str(k)
        if isinstance(v, dict):
            items.extend(flatten_dict(v, new_key, sep=sep, ignored_keys=ignored_keys).items())
        else:
            items.append((new_key, v))
    return dict(items)


def dictconfig_to_dict(cfg: TCfg) -> Dict[str, Any]:
    if isinstance(cfg, DictConfig):
        cfg = OmegaConf.to_container(cfg, resolve=True)

    ret = dict()

    for k in cfg.keys():

        if isinstance(cfg[k], DictConfig) or isinstance(cfg[k], dict):
            ret[k] = dictconfig_to_dict(cfg[k])
        else:
            ret[k] = cfg[k]

    return ret


def smart_sample(array: List[Any], k: int) -> List[Any]:
    """Sample n_samples items from given list. If array contains at least n_samples items, sample without repetition;
    otherwise take all the unique items and sample n_samples - len(array) ones with repetition.

    Args:
        array: list of unique elements to sample from
        k: number of items to sample

    Returns:
        sampled_items: list of sampled items
    """
    array_size = len(array)
    if array_size < k:
        sampled = (
            np.random.choice(array, size=array_size, replace=False).tolist()
            + np.random.choice(array, size=k - array_size, replace=True).tolist()
        )
    else:
        sampled = np.random.choice(array, size=k, replace=False).tolist()
    return sampled


def clip_max(arr: Tuple[int, ...], max_el: int) -> Tuple[int, ...]:
    return tuple(min(x, max_el) for x in arr)


def remove_unused_kwargs(kwargs: Dict[str, Any], function: Any) -> Dict[str, Any]:
    return {k: v for k, v in kwargs.items() if k in inspect.signature(function).parameters}


def check_if_nonempty_positive_integers(var: Union[int, Sequence[int]], name: str) -> None:
    """
    Check whether ``var`` is a positive integer or a non-empty Iterable of positive integers.

    Args:
        var: A sequence.
        name: A name of the sequence in case of exception should be raised.

    """
    if isinstance(var, Sequence):
        if not len(var) > 0 or not all([isinstance(x, int) and (x > 0) for x in var]):
            raise ValueError(f"{name} is expected to be non-empty and contain positive integers, but got {var}")
    elif isinstance(var, int):
        if var <= 0:
            raise ValueError(f"{name} is expected to be a positive integer, but got {var}")
    else:
        raise ValueError(f"Unsupported argument type. Expected int or Iterable[int], but got {type(var)}")


def compare_dicts_recursively(d1: Dict, d2: Dict) -> bool:  # type: ignore
    """
    The function compares dictionaries and prints the exact information where they differ. By using the built-in
    dictionary comparison one can get just a plain 'True' or 'False' as result of the comparison, without any hints
    on where the dictionaries differ.
    """
    assert set(d1.keys()) == set(
        d2.keys()
    ), f"The dictionaries keys are different.\nDict_1 keys: {set(d1.keys())}\nDict_2 keys: {set(d2.keys())}"
    for k, v in d1.items():
        if isinstance(v, dict):
            assert compare_dicts_recursively(
                v, d2[k]
            ), f"The dictionaries differs at key {k}.\nDict_1 value: {v}\nDict_2 value: {d2[k]}"
        elif isinstance(v, Tensor):
            assert torch.allclose(d2[k], v)
        else:
            assert d2[k] == v, f"Key name: {k}\nDict_1 value: {v}\nDict_2 value: {d2[k]}"
    return True


def pad_array_right(arr: np.ndarray, required_len: int, val: Union[float, int]) -> np.ndarray:
    assert required_len >= len(arr)
    assert arr.ndim == 1

    return np.pad(arr, (0, required_len - len(arr)), mode="constant", constant_values=val)


def visualise_text(text: str, color: TColor = BLACK, draw_bbox: bool = True) -> np.ndarray:
    fig, ax = plt.subplots(figsize=(2.56, 2.56), dpi=100)
    ax.text(0.5, 0.5, text, ha="center", va="center", wrap=True, fontsize=20)
    ax.set_xlim(0, 1)
    ax.set_ylim(0, 1)
    ax.axis("off")

    fig.canvas.draw()
    width, height = fig.canvas.get_width_height()
    image = np.frombuffer(fig.canvas.tostring_rgb(), dtype=np.uint8)
    image = image.reshape(height, width, 3)
    plt.close(fig)

    image = np.resize(image, (256, 256, 3))

    if draw_bbox:
        frame_thickness = 5
        image[:frame_thickness, :, :] = color  # Top frame
        image[-frame_thickness:, :, :] = color  # Bottom frame
        image[:, :frame_thickness, :] = color  # Left frame
        image[:, -frame_thickness:, :] = color  # Right frame

    return image


@contextmanager
def matplotlib_backend(backend: str) -> Generator[None, None, None]:
    current_backend = matplotlib.get_backend()
    try:
        matplotlib.use(backend)
        yield
    finally:
        matplotlib.use(current_backend)


__all__ = [
    "find_value_ids",
    "set_global_seed",
    "one_hot",
    "flatten_dict",
    "dictconfig_to_dict",
    "smart_sample",
    "clip_max",
    "check_if_nonempty_positive_integers",
    "compare_dicts_recursively",
    "pad_array_right",
    "visualise_text",
    "matplotlib_backend",
]



import itertools
import json
from pathlib import Path
from pprint import pprint

import pytorch_lightning as pl
from torch.utils.data import DataLoader

from oml.const import IMAGE_EXTENSIONS, TCfg
from oml.datasets.images import ImageBaseDataset
from oml.ddp.utils import get_world_size_safe, is_main_process, sync_dicts_ddp
from oml.lightning.modules.extractor import ExtractorModule
from oml.lightning.pipelines.parser import parse_engine_params_from_config
from oml.registry.models import get_extractor_by_cfg
from oml.registry.transforms import get_transforms_by_cfg
from oml.transforms.images.utils import get_im_reader_for_transforms
from oml.utils.images.images import find_broken_images
from oml.utils.misc import dictconfig_to_dict


def extractor_prediction_pipeline(cfg: TCfg) -> None:
    """
    This pipeline allows you to save features extracted by a feature extractor.

    The config can be specified as a dictionary or with hydra: https://hydra.cc/.
    For more details look at ``pipelines/features_extraction/README.md``

    """
    cfg = dictconfig_to_dict(cfg)

    pprint(cfg)

    transforms = get_transforms_by_cfg(cfg["transforms_predict"])
    filenames = [list(Path(cfg["data_dir"]).glob(f"**/*.{ext}")) for ext in IMAGE_EXTENSIONS]
    filenames = list(itertools.chain(*filenames))

    if len(filenames) == 0:
        raise RuntimeError(f"There are no images in the provided directory: {cfg['data_dir']}")

    f_imread = get_im_reader_for_transforms(transforms)

    print("Let's check if there are broken images:")
    broken_images = find_broken_images(filenames, f_imread=f_imread)
    if broken_images:
        raise ValueError(f"There are images that cannot be open:\n {broken_images}.")

    dataset = ImageBaseDataset(paths=filenames, transform=transforms, f_imread=f_imread)

    loader = DataLoader(
        dataset=dataset, batch_size=cfg["bs"], num_workers=cfg["num_workers"], shuffle=False, drop_last=False
    )

    extractor = get_extractor_by_cfg(cfg["extractor"])
    pl_model = ExtractorModule(extractor=extractor)

    trainer_engine_params = parse_engine_params_from_config(cfg)
    trainer_engine_params["use_distributed_sampler"] = True
    trainer = pl.Trainer(precision=cfg.get("precision", 32), **trainer_engine_params)
    predictions = trainer.predict(model=pl_model, dataloaders=loader, return_predictions=True)

    paths, embeddings = [], []
    for prediction in predictions:
        paths.extend([filenames[i] for i in prediction[dataset.index_key].tolist()])
        embeddings.extend(prediction[pl_model.embeddings_key].tolist())

    paths = sync_dicts_ddp({"key": list(map(str, paths))}, get_world_size_safe())["key"]
    embeddings = sync_dicts_ddp({"key": embeddings}, get_world_size_safe())["key"]

    if is_main_process():
        data_to_save = dict(zip(paths, embeddings))
        save_path = Path(cfg["save_dir"]) / "predictions.json"

        with open(save_path, "w") as f:
            json.dump(data_to_save, f)

        print(f"{len(paths)} predictions have been saved to {save_path}")


__all__ = ["extractor_prediction_pipeline"]



from typing import Any, Dict, List, Optional

import numpy as np
import torch
from torch import nn
from torch.nn import functional as F
from torchvision.ops import MLP

from oml.functional.label_smoothing import label_smoothing


class ArcFaceLoss(nn.Module):
    """
    ArcFace loss from `paper <https://arxiv.org/abs/1801.07698>`_ with possibility to use label smoothing.
    It contains projection size of ``num_features x num_classes`` inside itself. Please make sure that class labels
    started with 0 and ended as ``num_classes`` - 1.
    """

    criterion_name = "arcface"  # for better logging

    def __init__(
        self,
        in_features: int,
        num_classes: int,
        m: float = 0.5,
        s: float = 64,
        smoothing_epsilon: float = 0,
        label2category: Optional[Dict[Any, Any]] = None,
        reduction: str = "mean",
    ):
        """
        Args:
            in_features: Input feature size
            num_classes: Number of classes in train set
            m: Margin parameter for ArcFace loss. Usually you should use 0.3-0.5 values for it
            s: Scaling parameter for ArcFace loss. Usually you should use 30-64 values for it
            smoothing_epsilon: Label smoothing effect strength
            label2category: Optional, mapping from label to its category. If provided, label smoothing will redistribute
                 ``smoothing_epsilon`` only inside the category corresponding to the sample's ground truth label
            reduction: CrossEntropyLoss reduction
        """
        super(ArcFaceLoss, self).__init__()

        assert (
            smoothing_epsilon is None or 0 <= smoothing_epsilon < 1
        ), f"Choose another smoothing_epsilon parametrization, got {smoothing_epsilon}"

        self.criterion = nn.CrossEntropyLoss(reduction=reduction)
        self.num_classes = num_classes
        if label2category is not None:
            mapper = {l: i for i, l in enumerate(sorted(list(set(label2category.values()))))}
            label2category = {k: mapper[v] for k, v in label2category.items()}
            self.label2category = torch.arange(num_classes).apply_(label2category.get)
        else:
            self.label2category = None
        self.smoothing_epsilon = smoothing_epsilon
        self.weight = nn.Parameter(torch.FloatTensor(num_classes, in_features))
        nn.init.xavier_uniform_(self.weight)
        self.rescale = s
        self.m = m
        self.cos_m = np.cos(m)
        self.sin_m = np.sin(m)
        self.th = -self.cos_m
        self.mm = self.sin_m * m
        self._last_logs: Dict[str, float] = {}

    def fc(self, x: torch.Tensor) -> torch.Tensor:
        return F.linear(F.normalize(x, p=2), F.normalize(self.weight, p=2))

    def smooth_labels(self, y: torch.Tensor) -> torch.Tensor:
        if self.label2category is not None:
            self.label2category = self.label2category.to(self.weight.device)
        return label_smoothing(y, self.num_classes, self.smoothing_epsilon, self.label2category)

    def forward(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        assert torch.all(y < self.num_classes), "You should provide labels between 0 and num_classes - 1."

        cos = self.fc(x)

        self._log_accuracy_on_batch(cos, y)

        sin = torch.sqrt(1.0 - torch.pow(cos, 2))

        cos_w_margin = cos * self.cos_m - sin * self.sin_m
        cos_w_margin = torch.where(cos > self.th, cos_w_margin, cos - self.mm)

        ohe = F.one_hot(y, self.num_classes)
        logit = torch.where(ohe.bool(), cos_w_margin, cos) * self.rescale

        if self.smoothing_epsilon:
            y = self.smooth_labels(y)

        return self.criterion(logit, y)

    @torch.no_grad()
    def _log_accuracy_on_batch(self, logits: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        self._last_logs["accuracy"] = torch.mean((y == torch.argmax(logits, 1)).to(torch.float32))

    @property
    def last_logs(self) -> Dict[str, Any]:
        """
        Returns:
            Dictionary containing useful statistic calculated for the last batch.
        """
        return self._last_logs


class ArcFaceLossWithMLP(nn.Module):
    """
    Almost the same as ``ArcFaceLoss``, but also has MLP projector before the loss.
    You may want to use ``ArcFaceLossWithMLP`` to boost the expressive power of ArcFace loss during the training
    (for example, in a multi-head setup it may be a good idea to have task-specific projectors in each of the losses).
    Note, the criterion does not exist during the validation time.
    Thus, if you want to keep your MLP layers, you should create them as a part of the model you train.
    """

    def __init__(
        self,
        in_features: int,
        num_classes: int,
        mlp_features: List[int],
        m: float = 0.5,
        s: float = 64,
        smoothing_epsilon: float = 0,
        label2category: Optional[Dict[Any, Any]] = None,
        reduction: str = "mean",
    ):
        """
        Args:
            in_features: Input feature size
            num_classes: Number of classes in train set
            mlp_features: Layers sizes for MLP before ArcFace
            m: Margin parameter for ArcFace loss. Usually you should use 0.3-0.5 values for it
            s: Scaling parameter for ArcFace loss. Usually you should use 30-64 values for it
            smoothing_epsilon: Label smoothing effect strength
            label2category: Optional, mapping from label to its category. If provided, label smoothing will redistribute
                 ``smoothing_epsilon`` only inside the category corresponding to the sample's ground truth label
            reduction: CrossEntropyLoss reduction
        """
        super().__init__()
        self.mlp = MLP(in_features, mlp_features)
        self.arcface = ArcFaceLoss(
            mlp_features[-1],
            num_classes=num_classes,
            label2category=label2category,
            smoothing_epsilon=smoothing_epsilon,
            m=m,
            s=s,
            reduction=reduction,
        )

    def forward(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        return self.arcface(self.mlp(x), y)

    @property
    def last_logs(self) -> Dict[str, Any]:
        """
        Returns:
             Dictionary containing useful statistic calculated for the last batch.
        """
        return self.arcface.last_logs


__all__ = ["ArcFaceLoss", "ArcFaceLossWithMLP"]



from pathlib import Path
from typing import Any, Callable, Dict, List, Tuple

import torch
from torch import FloatTensor, Tensor, nn
from torch.utils.data import DataLoader
from tqdm.auto import tqdm

from oml.datasets import PairDataset
from oml.ddp.patching import patch_dataloader_to_ddp
from oml.ddp.utils import get_world_size_safe, is_ddp, sync_dicts_ddp
from oml.interfaces.datasets import IBaseDataset, IIndexedDataset
from oml.interfaces.models import IPairwiseModel
from oml.utils.misc_torch import get_device, temporary_setting_model_mode, unique_by_ids


@torch.no_grad()
def _inference(
    model: nn.Module,
    apply_model: Callable[[nn.Module, Dict[str, Any]], Tensor],
    dataset: IIndexedDataset,
    num_workers: int,
    batch_size: int,
    verbose: bool,
    use_fp16: bool,
    accumulate_on_cpu: bool = True,
) -> Tensor:
    loader = DataLoader(dataset=dataset, batch_size=batch_size, num_workers=num_workers, shuffle=False)

    if is_ddp():
        loader = patch_dataloader_to_ddp(loader)

    if verbose:
        loader = tqdm(loader, desc=str(get_device(model)))

    outputs_list = []
    ids = []

    with torch.autocast(device_type="cuda", dtype=torch.float16 if use_fp16 else torch.float32):
        with temporary_setting_model_mode(model, set_train=False):
            for batch in loader:
                out = apply_model(model, batch)
                if accumulate_on_cpu:
                    out = out.cpu()
                outputs_list.append(out)
                ids.extend(batch[dataset.index_key].long().tolist())

    outputs = torch.cat(outputs_list).detach()

    data_to_sync = {"outputs": outputs, "ids": ids}
    data_synced = sync_dicts_ddp(data_to_sync, world_size=get_world_size_safe())
    outputs, ids = data_synced["outputs"], data_synced["ids"]

    ids, outputs = unique_by_ids(ids=ids, data=outputs)

    assert len(outputs) == len(dataset), "Data was not collected correctly after DDP sync."
    assert list(range(len(dataset))) == ids, "Data was not collected correctly after DDP sync."

    return outputs


@torch.no_grad()
def inference(
    model: nn.Module,
    dataset: IBaseDataset,
    batch_size: int,
    num_workers: int = 0,
    verbose: bool = False,
    use_fp16: bool = False,
    accumulate_on_cpu: bool = True,
) -> Tensor:
    device = get_device(model)

    def apply(model_: nn.Module, batch_: Dict[str, Any]) -> FloatTensor:
        return model_(batch_[dataset.input_tensors_key].to(device))

    return _inference(
        model=model,
        apply_model=apply,
        dataset=dataset,
        num_workers=num_workers,
        batch_size=batch_size,
        verbose=verbose,
        use_fp16=use_fp16,
        accumulate_on_cpu=accumulate_on_cpu,
    )


def inference_cached(
    model: nn.Module,
    dataset: IBaseDataset,
    batch_size: int,
    num_workers: int = 0,
    use_fp16: bool = False,
    verbose: bool = True,
    accumulate_on_cpu: bool = True,
    cache_path: str = "inference_cache.pth",
) -> Tensor:
    if Path(cache_path).is_file():
        outputs = torch.load(cache_path, map_location="cpu")
        print(f"Model outputs have been loaded from {cache_path}.")
    else:
        outputs = inference(
            model=model,
            dataset=dataset,
            num_workers=num_workers,
            batch_size=batch_size,
            use_fp16=use_fp16,
            verbose=verbose,
            accumulate_on_cpu=accumulate_on_cpu,
        )

        torch.save(outputs, cache_path)
        print(f"Model outputs have been saved to {cache_path}.")

    return outputs


def pairwise_inference(
    model: IPairwiseModel,
    base_dataset: IBaseDataset,
    pair_ids: List[Tuple[int, int]],
    num_workers: int,
    batch_size: int,
    verbose: bool = True,
    use_fp16: bool = False,
    accumulate_on_cpu: bool = True,
) -> Tensor:
    device = get_device(model)

    dataset = PairDataset(base_dataset=base_dataset, pair_ids=pair_ids)

    def _apply(
        model_: IPairwiseModel,
        batch_: Dict[str, Any],
    ) -> Tensor:
        pair1 = batch_[dataset.input_tensors_key_1].to(device)
        pair2 = batch_[dataset.input_tensors_key_2].to(device)
        return model_.predict(pair1, pair2)

    output = _inference(
        model=model,
        apply_model=_apply,
        dataset=dataset,
        num_workers=num_workers,
        batch_size=batch_size,
        verbose=verbose,
        use_fp16=use_fp16,
        accumulate_on_cpu=accumulate_on_cpu,
    )

    return output


__all__ = ["inference", "pairwise_inference", "inference_cached"]



from typing import Sequence, Tuple

from torch import FloatTensor, LongTensor, concat, finfo

from oml.inference.abstract import pairwise_inference
from oml.interfaces.datasets import IQueryGalleryDataset
from oml.interfaces.models import IPairwiseModel
from oml.interfaces.retrieval import IRetrievalPostprocessor
from oml.retrieval.retrieval_results import RetrievalResults
from oml.utils.misc_torch import cat_two_sorted_tensors_and_keep_it_sorted


class PairwiseReranker(IRetrievalPostprocessor):
    def __init__(
        self,
        top_n: int,
        pairwise_model: IPairwiseModel,
        num_workers: int,
        batch_size: int,
        verbose: bool = False,
        use_fp16: bool = False,
    ):
        """

        Args:
            top_n: Model will be applied to the ``num_queries * top_n`` pairs formed by each query
                and ``top_n`` most relevant galleries.
            pairwise_model: Model which is able to take two items as inputs
                and estimate the *distance* (not in a strictly mathematical sense) between them.
            num_workers: Number of workers in DataLoader
            batch_size: Batch size that will be used in DataLoader
            verbose: Set ``True`` if you want to see progress bar for an inference
            use_fp16: Set ``True`` if you want to use half precision

        """
        assert top_n > 1, "The number of the retrieved results for each query to process has to be greater than 1."

        self._top_n = top_n
        self.model = pairwise_model

        self.num_workers = num_workers
        self.batch_size = batch_size
        self.verbose = verbose
        self.use_fp16 = use_fp16

    @property
    def top_n(self) -> int:
        """
        Returns:
            Number of gallery items closest to each query to process.
        """
        return self._top_n

    def process(self, rr: RetrievalResults, dataset: IQueryGalleryDataset) -> RetrievalResults:  # type: ignore
        """

        Args:
            rr: ``RetrievalResults`` object.
            dataset: Dataset having query/gallery split.

        Returns:
            After re-ranking is applied to the ``top_n`` retrieved items, the updated ``RetrievalResults`` are returned.
            In other words, we permute the first ``top_n`` items, but the rest remains untouched.

        **Example 1** (for one query):

        .. code-block:: python

            rr.retrieved_ids = [[3,   2,   1,   0,   4  ]]
            rr.distances     = [[0.1, 0.2, 0.5, 0.6, 0.7]]

            # Let's say a postprocessor has been applied to the
            # first 3 elements and the new distances are: [0.4, 0.2, 0.3]

            # In this case, the updated values will be:
            rr.retrieved_ids = [[2,   1,   3,   0,   4  ]]
            rr.distances:    = [[0.2, 0.3, 0.4, 0.6, 0.7]]

        **Example 2** (for one query):

        .. code-block:: python

            # Note, the new distances to the top_n items produced by the pairwise model
            # may be rescaled to keep the distances order. Here is an example:
            rr.distances = [[0.1, 0.2, 0.3, 0.5, 0.6]]
            top_n = 3

            # Imagine, the postprocessor didn't change the order of the first 3 items
            # (it's just a convenient example, the general logic remains the same),
            # however the new values have a bigger scale:
            distances_new = [[1, 2, 5, 0.5, 0.6]]

            # Thus, we need to downscale the first 3 distances, so they are lower than 0.5:
            scale = 5 / 0.5 = 0.1
            # Finally, let's apply the found scale to the top 3 distances:
            rr_upd.distances = [[0.1, 0.2, 0.5, 0.5, 0.6]]

            # Note, if new and old distances are already sorted, we don't apply any scaling.

        """
        assert len(dataset.get_query_ids()) == len(
            rr.retrieved_ids
        ), f"{rr.__class__.__name__} and {dataset.__class__.__name__} must have the same number of queries."

        gt_ids = rr.gt_ids
        distances_upd, retrieved_ids_upd = self._process_raw(
            retrieved_ids=rr.retrieved_ids, distances=rr.distances, dataset=dataset
        )
        rr_upd = RetrievalResults(distances=distances_upd, retrieved_ids=retrieved_ids_upd, gt_ids=gt_ids)
        return rr_upd

    def _process_raw(
        self, retrieved_ids: Sequence[LongTensor], distances: Sequence[FloatTensor], dataset: IQueryGalleryDataset
    ) -> Tuple[Sequence[FloatTensor], Sequence[LongTensor]]:

        # Let's make list of pairs of queries and top_n gallery items for which we need to recompute distances
        # Queries have different number of retrieved items, so we track what pairs are relevant to what queries (bounds)
        pairs = []
        bounds = [0]
        query_ids = dataset.get_query_ids()
        gallery_ids = dataset.get_gallery_ids()
        for iq, ids_gallery in enumerate(retrieved_ids):
            ids_gallery_global = gallery_ids[ids_gallery][: self.top_n].tolist()
            ids_query_global = [query_ids[iq].item()] * len(ids_gallery_global)

            pairs.extend(list(zip(ids_query_global, ids_gallery_global)))
            bounds.append(bounds[-1] + len(ids_gallery_global))

        distances_recomputed = pairwise_inference(
            model=self.model,
            base_dataset=dataset,
            pair_ids=pairs,
            num_workers=self.num_workers,
            batch_size=self.batch_size,
            verbose=self.verbose,
            use_fp16=self.use_fp16,
        )

        # Reshape flat dists into the original structure of sequences (having different sizes) relevant to each query
        distances_upd, retrieved_ids_upd = [], []
        for query_start, query_end, dist_orig, ri_orig in zip(bounds[:-1], bounds[1:], distances, retrieved_ids):
            dist_recomputed_q, ii_rerank = distances_recomputed[query_start:query_end].sort()

            distances_upd += [
                cat_two_sorted_tensors_and_keep_it_sorted(
                    dist_recomputed_q.view(1, -1),
                    dist_orig[self.top_n :].view(1, -1),
                    eps=10**3 * finfo(distances_recomputed.dtype).eps,
                ).view(-1)
            ]
            retrieved_ids_upd += [concat([ri_orig[ii_rerank], ri_orig[self.top_n :]])]

        for iq in range(len(retrieved_ids)):
            # Re-ranking cannot change the number of retrieved items. It may only change their order.
            assert len(retrieved_ids[iq]) == len(retrieved_ids_upd[iq])
            assert len(distances[iq]) == len(distances_upd[iq])

        return distances_upd, retrieved_ids_upd


__all__ = ["PairwiseReranker"]



from typing import Dict, List, Tuple, Union

from torch import Tensor

from oml.const import INDEX_KEY, INPUT_TENSORS_KEY_1, INPUT_TENSORS_KEY_2
from oml.interfaces.datasets import IBaseDataset, IPairDataset


class PairDataset(IPairDataset):
    """
    Dataset to iterate over pairs of items of any modality.

    """

    def __init__(
        self,
        base_dataset: IBaseDataset,
        pair_ids: List[Tuple[int, int]],
        input_tensors_key_1: str = INPUT_TENSORS_KEY_1,
        input_tensors_key_2: str = INPUT_TENSORS_KEY_2,
        index_key: str = INDEX_KEY,
    ):
        self.base_dataset = base_dataset
        self.pair_ids = pair_ids

        self.input_tensors_key_1 = input_tensors_key_1
        self.input_tensors_key_2 = input_tensors_key_2
        self.index_key: str = index_key

    def __getitem__(self, item: int) -> Dict[str, Union[Tensor, int]]:
        i1, i2 = self.pair_ids[item]
        key = self.base_dataset.input_tensors_key
        return {
            self.input_tensors_key_1: self.base_dataset[i1][key],
            self.input_tensors_key_2: self.base_dataset[i2][key],
            self.index_key: item,
        }

    def __len__(self) -> int:
        return len(self.pair_ids)


__all__ = ["PairDataset"]



from typing import Any, Dict, Optional, Union

import pytorch_lightning as pl
import torch
from pytorch_lightning.utilities.types import EVAL_DATALOADERS, TRAIN_DATALOADERS
from torch import nn
from torch.optim.lr_scheduler import ReduceLROnPlateau, _LRScheduler

from oml.const import EMBEDDINGS_KEY, INPUT_TENSORS_KEY, LABELS_KEY
from oml.interfaces.models import IExtractor, IFreezable
from oml.lightning.modules.ddp import ModuleDDP


class ExtractorModule(pl.LightningModule):
    """
    This is a base module to train your model with Lightning.

    """

    def __init__(
        self,
        extractor: IExtractor,
        criterion: Optional[nn.Module] = None,
        optimizer: Optional[torch.optim.Optimizer] = None,
        scheduler: Optional[_LRScheduler] = None,
        scheduler_interval: str = "step",
        scheduler_frequency: int = 1,
        input_tensors_key: str = INPUT_TENSORS_KEY,
        labels_key: str = LABELS_KEY,
        embeddings_key: str = EMBEDDINGS_KEY,
        scheduler_monitor_metric: Optional[str] = None,
        freeze_n_epochs: int = 0,
    ):
        """

        Args:
            extractor: Extractor to train
            criterion: Criterion to optimize
            optimizer: Optimizer
            scheduler: Learning rate scheduler
            scheduler_interval: Interval of calling scheduler (must be ``step`` or ``epoch``)
            scheduler_frequency: Frequency of calling scheduler
            input_tensors_key: Key to get tensors from the batches
            labels_key: Key to get labels from the batches
            embeddings_key: Key to get embeddings from the batches
            scheduler_monitor_metric: Metric to monitor for the schedulers that depend on the metric value
            freeze_n_epochs: number of epochs to freeze model (for n > 0 model has to be a successor of ``IFreezable``
                interface). When ``current_epoch >= freeze_n_epochs`` model is unfreezed. Note that epochs are
                starting with 0.

        """
        pl.LightningModule.__init__(self)

        self.model = extractor
        self.criterion = criterion
        self.optimizer = optimizer

        self.monitor_metric = scheduler_monitor_metric
        self.scheduler = scheduler
        self.scheduler_interval = scheduler_interval
        self.scheduler_frequency = scheduler_frequency

        self.input_tensors_key = input_tensors_key
        self.labels_key = labels_key
        self.embeddings_key = embeddings_key

        self.freeze_n_epochs = freeze_n_epochs
        assert freeze_n_epochs == 0 or isinstance(
            extractor, IFreezable
        ), f"Model must be {IFreezable.__name__} to use this."

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        embeddings = self.model(x)
        return embeddings

    def training_step(self, batch: Dict[str, Any], batch_idx: int) -> torch.Tensor:
        embeddings = self.model(batch[self.input_tensors_key])
        bs = len(embeddings)

        loss = self.criterion(embeddings, batch[self.labels_key])
        loss_name = (getattr(self.criterion, "criterion_name", "") + "_loss").strip("_")
        self.log(loss_name, loss.item(), prog_bar=True, batch_size=bs, on_step=True, on_epoch=True)

        if hasattr(self.criterion, "last_logs"):
            self.log_dict(self.criterion.last_logs, prog_bar=False, batch_size=bs, on_step=True, on_epoch=False)

        if self.scheduler is not None:
            self.log("lr", self.scheduler.get_last_lr()[0], prog_bar=True, batch_size=bs, on_step=True, on_epoch=False)

        return loss

    def validation_step(self, batch: Dict[str, Any], batch_idx: int, *_: Any) -> Dict[str, Any]:
        embeddings = self.model.extract(batch[self.input_tensors_key])
        return {**batch, **{self.embeddings_key: embeddings}}

    def predict_step(self, batch: Dict[str, Any], batch_idx: int, *_: Any) -> Dict[str, Any]:
        return self.validation_step(batch, batch_idx)

    def configure_optimizers(self) -> Any:
        if self.scheduler is None:
            return self.optimizer
        else:
            scheduler = {
                "scheduler": self.scheduler,
                "interval": self.scheduler_interval,
                "frequency": self.scheduler_frequency,
            }
            if isinstance(self.scheduler, ReduceLROnPlateau):
                scheduler["monitor"] = self.monitor_metric
            return [self.optimizer], [scheduler]

    def get_progress_bar_dict(self) -> Dict[str, Union[int, str]]:
        # https://github.com/Lightning-AI/lightning/issues/1595
        tqdm_dict = super().get_progress_bar_dict()
        tqdm_dict.pop("v_num", None)
        return tqdm_dict

    def on_epoch_start(self) -> None:
        if self.freeze_n_epochs > 0:
            if not isinstance(self.model, IFreezable):
                raise ValueError(
                    f"You want to freeze a model for {self.freeze_n_epochs},"
                    f"but model does not implement the {IFreezable.__name__} interface."
                )

            if self.current_epoch >= self.freeze_n_epochs:
                self.model.unfreeze()
            else:
                self.model.freeze()


class ExtractorModuleDDP(ExtractorModule, ModuleDDP):
    """
    This is a base module for the training of your model with Lightning in DDP.

    """

    def __init__(
        self,
        loaders_train: Optional[TRAIN_DATALOADERS] = None,
        loaders_val: Optional[EVAL_DATALOADERS] = None,
        *args: Any,
        **kwargs: Any,
    ):
        ModuleDDP.__init__(self, loaders_train=loaders_train, loaders_val=loaders_val)
        ExtractorModule.__init__(self, *args, **kwargs)


__all__ = ["ExtractorModule", "ExtractorModuleDDP"]



from pathlib import Path
from typing import Any, Dict, Optional

import torch
from pytorch_lightning.callbacks import ModelCheckpoint
from pytorch_lightning.strategies import DDPStrategy

from oml.const import CATEGORIES_COLUMN, TCfg
from oml.interfaces.datasets import ILabeledDataset
from oml.interfaces.loggers import IPipelineLogger
from oml.interfaces.samplers import IBatchSampler
from oml.lightning.pipelines.logging import TensorBoardPipelineLogger
from oml.registry.loggers import get_logger_by_cfg
from oml.registry.samplers import SAMPLERS_CATEGORIES_BASED, get_sampler_by_cfg
from oml.registry.schedulers import get_scheduler_by_cfg
from oml.utils.misc import dictconfig_to_dict


def parse_engine_params_from_config(cfg: TCfg) -> Dict[str, Any]:
    """
    The function parses config and based on it prepares DDP parameters for PytorchLightning Trainer module.
    There are two parameters 'accelerator' and 'devices' you can configure. If one of the parameters or both of them
    are not specified, the best option based on your hardware will be automatically prepared for you.
    Possible options for 'accelerator' are 'cpu' and 'gpu'.
    You can select specific GPUs using the list 'devices=[0, 2]' or specify the number of GPUs by `devices=3`.
    An experiment might be launched in DDP with the 'cpu' accelerator. In this case, 'devices' (integer value or
    length of list) interpreted as a number of processes.
    """
    cfg = dictconfig_to_dict(cfg)

    # we want to replace possible null or no values in config for "accelerator" and "devices"
    accelerator = cfg.get("accelerator")
    if accelerator is None:
        accelerator = "gpu" if torch.cuda.is_available() else "cpu"

    devices = cfg.get("devices")
    if devices is None:
        devices = torch.cuda.device_count() if (torch.cuda.is_available() and accelerator == "gpu") else 1

    if isinstance(devices, (list, tuple)) and accelerator == "cpu":
        devices = len(devices)

    if (isinstance(devices, int) and devices > 1) or (isinstance(devices, (list, tuple)) and len(devices) > 1):
        strategy = DDPStrategy(find_unused_parameters=cfg.get("find_unused_parameters", False))
    else:
        strategy = "auto"

    return {
        "devices": devices,
        "strategy": strategy,
        "accelerator": accelerator,
        "use_distributed_sampler": False,
    }


def check_is_config_for_ddp(cfg: TCfg) -> bool:
    return bool(cfg["strategy"])


def parse_logger_from_config(cfg: TCfg) -> IPipelineLogger:
    logger = TensorBoardPipelineLogger(".") if cfg.get("logger", None) is None else get_logger_by_cfg(cfg["logger"])
    return logger  # type: ignore


def parse_scheduler_from_config(cfg: TCfg, optimizer: torch.optim.Optimizer) -> Dict[str, Any]:
    if cfg.get("scheduling"):
        scheduler_kwargs = {
            "scheduler": get_scheduler_by_cfg(cfg["scheduling"]["scheduler"], **{"optimizer": optimizer}),
            "scheduler_interval": cfg["scheduling"]["scheduler_interval"],
            "scheduler_frequency": cfg["scheduling"]["scheduler_frequency"],
            "scheduler_monitor_metric": cfg["scheduling"].get("monitor_metric", None),
        }
    else:
        scheduler_kwargs = {"scheduler": None}

    return scheduler_kwargs


def parse_sampler_from_config(cfg: TCfg, dataset: ILabeledDataset) -> Optional[IBatchSampler]:
    if (
        (CATEGORIES_COLUMN not in dataset.extra_data)
        and (cfg["sampler"] is not None)
        and cfg["sampler"]["name"] in SAMPLERS_CATEGORIES_BASED.keys()
    ):
        raise ValueError(
            "NOTE! You are trying to use Sampler which works with the information related"
            "to categories, but there is no <categories_key> in your Dataset."
        )

    sampler_runtime_args = {"labels": dataset.get_labels(), "label2category": dataset.get_label2category()}
    sampler = get_sampler_by_cfg(cfg["sampler"], **sampler_runtime_args) if cfg["sampler"] is not None else None

    return sampler


def parse_ckpt_callback_from_config(cfg: TCfg) -> ModelCheckpoint:
    return ModelCheckpoint(
        dirpath=Path.cwd() / "checkpoints",
        monitor=cfg["metric_for_checkpointing"],
        mode=cfg.get("mode_for_checkpointing", "max"),
        save_top_k=1,
        verbose=True,
        filename="best",
    )


__all__ = [
    "parse_engine_params_from_config",
    "check_is_config_for_ddp",
    "parse_scheduler_from_config",
    "parse_sampler_from_config",
    "parse_ckpt_callback_from_config",
]



# type: ignore
# flake8: noqa

import hashlib
import os
import urllib
import warnings
from typing import List

import torch
from tqdm import tqdm

from oml.models.vit_unicom.external.vision_transformer import load_model_and_transform

# ============== CODE FROM UNICOM ==============
# https://github.com/deepglint/unicom/blob/main/unicom/model.py

_MODELS = {
    "ViT-B/32": "https://github.com/deepglint/unicom/releases/download/b32/FP16-ViT-B-32.pt",
    "ViT-B/16": "https://github.com/deepglint/unicom/releases/download/b16/FP16-ViT-B-16.pt",
    "ViT-L/14": "https://github.com/deepglint/unicom/releases/download/l14/FP16-ViT-L-14.pt",
    "ViT-L/14@336px": "https://github.com/deepglint/unicom/releases/download/l14_336px/FP16-ViT-L-14-336px.pt",
}

_SHA256 = {
    "FP16-ViT-B-32.pt": "f9d5696a9b58dbbbefee2d31615ca59084f2895a0fdd2ca4c235e0f9b2793f7a",
    "FP16-ViT-B-16.pt": "c04f324f7c3b4435667236ec6c0eca1cd62f9d64fbfc2d06f8e8e60e6497edef",
    "FP16-ViT-L-14.pt": "ff3ab62ff782876460099e6e0ee17b73a7c01109de2fffd595f16f4129404bbd",
    "FP16-ViT-L-14-336px.pt": "3916ab5aed3b522fc90345be8b4457fe5dad60801ad2af5a6871c0c096e8d7ea",
}


def available_models() -> List[str]:
    """Returns the names of available CLIP models"""
    return list(_MODELS.keys())


def rm_module_from_state_dict(state_dict: dict) -> dict:
    result = {}
    for k, value in state_dict.items():

        if "module." in k:
            k_removed = k.split("module.")[-1]
            result[k_removed] = value
        else:
            result[k] = value
    return result


# copy from https://github.com/openai/CLIP/blob/main/clip/clip.py#L43
def _download(url: str, root: str):
    os.makedirs(root, exist_ok=True)
    filename = os.path.basename(url)

    expected_sha256 = _SHA256[filename]
    download_target = os.path.join(root, filename)

    if os.path.exists(download_target) and not os.path.isfile(download_target):
        raise RuntimeError(f"{download_target} exists and is not a regular file")

    if os.path.isfile(download_target):
        if hashlib.sha256(open(download_target, "rb").read()).hexdigest() == expected_sha256:
            return download_target
        else:
            warnings.warn(f"{download_target} exists, but the SHA256 checksum does not match; re-downloading the file")

    with urllib.request.urlopen(url) as source, open(download_target, "wb") as output:
        with tqdm(
            total=int(source.info().get("Content-Length")), ncols=80, unit="iB", unit_scale=True, unit_divisor=1024
        ) as loop:
            while True:
                buffer = source.read(8192)
                if not buffer:
                    break

                output.write(buffer)
                loop.update(len(buffer))

    if hashlib.sha256(open(download_target, "rb").read()).hexdigest() != expected_sha256:
        raise RuntimeError("Model has been downloaded but the SHA256 checksum does not not match")

    return download_target


# copy from https://github.com/openai/CLIP/blob/main/clip/clip.py#L94
def load(name: str, device: str = "cpu", download_root: str = None):
    if name in _MODELS:
        model_path = _download(_MODELS[name], download_root or os.path.expanduser("~/.cache/unicom"))
    elif os.path.isfile(name):
        model_path = name
    else:
        raise RuntimeError(f"Model {name} not found; available models = {available_models()}")
    with open(model_path, "rb") as opened_file:
        state_dict = torch.load(opened_file, map_location="cpu")

    model, transform = load_model_and_transform(name)
    state_dict_fp32 = {}
    for k, v in state_dict.items():
        state_dict_fp32[k] = v.float()

    model.load_state_dict(state_dict)
    return model, transform



from collections import defaultdict
from typing import Any, Dict, List, Optional, Tuple

import torch
from torch import Tensor, no_grad

from oml.interfaces.miners import ITripletsMiner
from oml.miners.inbatch_nhard_tri import NHardTripletsMiner
from oml.utils.misc_torch import OnlineAvgDict


class MinerWithBank(ITripletsMiner):
    """
    This is a class for cross-batch memory. This implementation uses only samples from the current batch as anchors
    and finds positive and negative pairs from the bank and the current batch using miner.

    """

    def __init__(
        self,
        bank_size_in_batches: int,
        miner: NHardTripletsMiner,
        need_logs: bool = True,
    ):
        """

        Args:
            bank_size_in_batches: Size of the bank.
            miner: Miner, for now we only support ``NHardTripletsMiner``
            need_logs: Set ``True`` to store some information to track in ``self.last_logs`` property.

        """

        assert isinstance(bank_size_in_batches, int)
        assert bank_size_in_batches >= 1
        assert isinstance(miner, NHardTripletsMiner)
        self.miner = miner

        self.bank_size_in_batches = bank_size_in_batches
        self.bank_features: Optional[Tensor] = None
        self.bank_labels: Optional[Tensor] = None
        self.bank_size = -1  # bank size in the number of vectors
        self.is_accumulated = False  # indicates if we filled out the bank
        # We maintain a queue of feature vectors and ptr indicates the place to insert another batch.
        # After the batch has been inserted, we move the pointer.
        self.ptr = 0

        self.need_logs = need_logs
        self._last_logs: Dict[str, float] = {}

    @no_grad()
    def __allocate_if_needed(self, features: Tensor, labels: Tensor) -> None:
        if self.bank_features is None:
            assert len(features) == len(labels)

            bs = features.shape[0]
            self.feat_dim = features.shape[-1]
            self.bank_size = self.bank_size_in_batches * bs
            self.bank_labels = torch.empty(self.bank_size, dtype=torch.long, device=features.device)
            self.bank_features = torch.empty(
                (self.bank_size, self.feat_dim), dtype=features.dtype, device=features.device
            )

    @no_grad()
    def _update_bank(self, features: Tensor, labels: Tensor) -> None:
        bs = features.shape[0]
        if not self.is_accumulated:
            self.is_accumulated = self.ptr + bs >= self.bank_size

        self.bank_features[self.ptr : self.ptr + bs] = features.clone().detach()
        self.bank_labels[self.ptr : self.ptr + bs] = labels.clone()
        self.ptr = (self.ptr + bs) % self.bank_size

    def sample(self, features: Tensor, labels: Tensor) -> Tuple[Tensor, Tensor, Tensor]:
        """

        Args:
            features: Features with the shape ``[batch_size, features_dim]``
            labels: Labels with the size of ``batch_size``

        Returns:
            Batch of triplets in the following order: anchor, positive, negative

        """
        if isinstance(labels, (list, tuple)):
            labels = torch.tensor(labels, dtype=torch.long, device=features.device)

        self.__allocate_if_needed(features=features, labels=labels)

        if self.is_accumulated:
            len_batch = len(labels)
            features_miner = torch.cat([features, self.bank_features], dim=0)
            labels_miner = torch.cat([labels, self.bank_labels], dim=0)
            ignore_anchor_mask = torch.zeros(len(labels_miner), dtype=torch.bool, device=features.device)
            ignore_anchor_mask[len_batch:] = True
        else:
            features_miner = features
            labels_miner = labels
            ignore_anchor_mask = torch.zeros(len(labels_miner), dtype=torch.bool, device=features.device)

        ids_a, ids_p, ids_n = self.miner._sample(
            features=features_miner, labels=labels_miner, ignore_anchor_mask=ignore_anchor_mask
        )

        if self.need_logs:
            self._last_logs = self._prepare_logs(
                ids_a=ids_a, ids_p=ids_p, ids_n=ids_n, ignore_anchor_mask=ignore_anchor_mask
            )

        self._update_bank(features=features, labels=labels)

        return features_miner[ids_a], features_miner[ids_p], features_miner[ids_n]

    @staticmethod
    def _prepare_logs(
        ids_a: List[int], ids_p: List[int], ids_n: List[int], ignore_anchor_mask: Tensor
    ) -> Dict[str, float]:
        logs = OnlineAvgDict()

        unq_triplets = set(zip(ids_a, ids_p, ids_n))
        ids_anchor2positives = defaultdict(set)
        ids_anchor2negatives = defaultdict(set)

        for anch, pos, neg in unq_triplets:
            ids_anchor2positives[anch].add(int(pos))
            ids_anchor2negatives[anch].add(int(neg))

        for anch in ids_anchor2positives.keys():
            positives = ids_anchor2positives[anch]
            positives_from_bank = ignore_anchor_mask[list(positives)].sum().item()
            positives_from_batch = len(positives) - positives_from_bank
            logs.update({"positives_from_bank": positives_from_bank, "positives_from_batch": positives_from_batch})

            negatives = ids_anchor2negatives[anch]
            negatives_from_bank = ignore_anchor_mask[list(negatives)].sum().item()
            negatives_from_batch = len(negatives) - negatives_from_bank
            logs.update({"negatives_from_bank": negatives_from_bank, "negatives_from_batch": negatives_from_batch})

        return logs.get_dict_with_results()

    @property
    def last_logs(self) -> Dict[str, Any]:
        """
        Returns:
            Dictionary containing useful statistic calculated for the last batch.
        """
        return self._last_logs


__all__ = ["MinerWithBank"]



from pathlib import Path
from typing import Any, Dict, Iterable, Optional

import torch

from oml.interfaces.models import IExtractor
from oml.models.utils import (
    TStateDict,
    filter_state_dict,
    patch_device_and_float,
    remove_criterion_in_state_dict,
    remove_prefix_from_state_dict,
)
from oml.models.vit_clip.external.model import VisionTransformer
from oml.utils.io import download_checkpoint

_OPENAI_URL = "https://openaipublic.azureedge.net/clip/models"
_SBER_URL = "https://huggingface.co/sberbank-ai"


def vitb16_224() -> VisionTransformer:
    return VisionTransformer(
        output_dim=512,
        input_resolution=224,
        layers=12,
        width=768,
        patch_size=16,
        heads=12,
    )


def vitb32_224() -> VisionTransformer:
    return VisionTransformer(
        output_dim=512,
        input_resolution=224,
        layers=12,
        width=768,
        patch_size=32,
        heads=12,
    )


def vitl14_224() -> VisionTransformer:
    return VisionTransformer(
        output_dim=768,
        input_resolution=224,
        layers=24,
        width=1024,
        patch_size=14,
        heads=16,
    )


def vitl14_336() -> VisionTransformer:
    return VisionTransformer(
        output_dim=768,
        input_resolution=224,
        layers=24,
        width=1024,
        patch_size=14,
        heads=16,
    )


class ViTCLIPExtractor(IExtractor):
    constructors = {
        "vitb16_224": vitb16_224,
        "vitb32_224": vitb32_224,
        "vitl14_224": vitl14_224,
        "vitl14_336": vitl14_336,
    }

    pretrained_models: Dict[str, Any] = {
        # checkpoints pretrained by OpenAI
        "openai_vitb16_224": {
            "url": f"{_OPENAI_URL}/5806e77cd80f8b59890b7e101eabd078d9fb84e6937f9e85e4ecb61988df416f/ViT-B-16.pt",
            "hash": "44c3d804ecac03d9545ac1a3adbca3a6",
            "is_jitted": True,
            "fname": "openai_vitb16_224.ckpt",
            "init_args": {"arch": "vitb16_224", "normalise_features": False},
        },
        "openai_vitb32_224": {
            "url": f"{_OPENAI_URL}/40d365715913c9da98579312b702a82c18be219cc2a73407c4526f58eba950af/ViT-B-32.pt",
            "hash": "3ba34e387b24dfe590eeb1ae6a8a122b",
            "is_jitted": True,
            "fname": "openai_vitb32_224.ckpt",
            "init_args": {"arch": "vitb32_224", "normalise_features": False},
        },
        "openai_vitl14_224": {
            "url": f"{_OPENAI_URL}/b8cca3fd41ae0c99ba7e8951adf17d267cdb84cd88be6f7c2e0eca1737a03836/ViT-L-14.pt",
            "hash": "096db1af569b284eb76b3881534822d9",
            "is_jitted": True,
            "fname": "openai_vitl14_224.ckpt",
            "init_args": {"arch": "vitl14_224", "normalise_features": False},
        },
        # checkpoints pretrained by SberbankAI
        "sber_vitb16_224": {
            "url": f"{_SBER_URL}/ruclip-vit-base-patch16-224/resolve/main/pytorch_model.bin",
            "hash": "7882e07674d78c674e33cb892a68bbfc",
            "is_jitted": False,
            "fname": "sber_vitb16_224.ckpt",
            "init_args": {"arch": "vitb16_224", "normalise_features": False},
        },
        "sber_vitb32_224": {
            "url": f"{_SBER_URL}/ruclip-vit-base-patch32-224/resolve/main/pytorch_model.bin",
            "hash": "e2c4dab46a3cfa608bdd762973e90d32",
            "is_jitted": False,
            "fname": "sber_vitb32_224.ckpt",
            "init_args": {"arch": "vitb32_224", "normalise_features": False},
        },
        "sber_vitl14_224": {
            "url": f"{_SBER_URL}/ruclip-vit-large-patch14-224/resolve/main/pytorch_model.bin",
            "hash": "9b4a1cd25d15bad4ffd2ba6e34b8a67c",
            "is_jitted": False,
            "fname": "sber_vitl14_224.ckpt",
            "init_args": {"arch": "vitl14_224", "normalise_features": False},
        },
    }

    def __init__(
        self,
        weights: Optional[str],
        arch: str,
        normalise_features: bool = True,
    ):
        """
        Args:
            weights: Path to weights or special key for pretrained ones or ``None`` for random initialization.
             You can check available pretrained checkpoints in ``ViTCLIPExtractor.pretrained_models``.
            arch: Might be one of ``vitb16_224``, ``vitb32_224``, ``vitl14_224``, ``vitl14_336``.
            normalise_features: Set ``True`` to normalise output features
        """

        super().__init__()

        self.normalize = normalise_features
        self.visual = self.constructors[arch]()

        self.input_size = int(arch.split("_")[-1])

        if weights is None:
            return
        if weights in self.pretrained_models:
            pretrained = self.pretrained_models[weights]
            jitted_weights = pretrained["is_jitted"]
            weights = download_checkpoint(pretrained["url"], pretrained["hash"], fname=pretrained["fname"])
        else:
            jitted_weights = False

        if jitted_weights:  # check if weights are jitted
            visual = torch.jit.load(Path(weights), map_location="cpu").visual
            patch_device_and_float(visual, device="cpu")
            state_dict = visual.state_dict()

        else:
            state_dict = torch.load(Path(weights), map_location="cpu")
            state_dict = state_dict.get("state_dict", state_dict)
            state_dict = remove_criterion_in_state_dict(state_dict)
            state_dict = take_visual_part_of_vit_clip(state_dict, needed_keys=self.visual.state_dict().keys())
            state_dict = remove_prefix_from_state_dict(state_dict, trial_key="conv1.weight")

        self.visual.load_state_dict(state_dict=state_dict, strict=True)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        assert (x.shape[-2] == self.input_size) and (
            x.shape[-1] == self.input_size
        ), f"The model expects input images to be resized to {self.input_size}x{self.input_size}"

        res = self.visual.forward(x)
        if self.normalize:
            res = res / torch.linalg.norm(res, 2, dim=1, keepdim=True).detach()
        return res

    @property
    def feat_dim(self) -> int:
        return self.visual.state_dict()["proj"].shape[-1]


def take_visual_part_of_vit_clip(state_dict: TStateDict, needed_keys: Iterable[str]) -> TStateDict:
    for k in list(state_dict):
        if "visual" in k:
            new_key = k[k.find("visual") + len("visual") + 1 :]
            state_dict[new_key] = state_dict.pop(k)
    state_dict = filter_state_dict(state_dict, needed_keys=needed_keys)
    return state_dict


__all__ = ["ViTCLIPExtractor"]



from typing import Any, Dict, Optional, Union

import pytorch_lightning as pl
from pytorch_lightning.utilities.types import EVAL_DATALOADERS, TRAIN_DATALOADERS
from torch import Tensor, nn
from torch.optim import Optimizer
from torch.optim.lr_scheduler import ReduceLROnPlateau, _LRScheduler

from oml.const import EMBEDDINGS_KEY, INPUT_TENSORS_KEY, LABELS_KEY
from oml.interfaces.models import IFreezable, IPairwiseModel
from oml.lightning.modules.ddp import ModuleDDP
from oml.miners.pairs import PairsMiner


class PairwiseModule(pl.LightningModule):
    def __init__(
        self,
        pairwise_model: IPairwiseModel,
        pairs_miner: PairsMiner,
        criterion: nn.Module,
        optimizer: Optimizer,
        scheduler: Optional[_LRScheduler] = None,
        scheduler_interval: str = "step",
        scheduler_frequency: int = 1,
        input_tensors_key: str = INPUT_TENSORS_KEY,
        labels_key: str = LABELS_KEY,
        embeddings_key: str = EMBEDDINGS_KEY,
        scheduler_monitor_metric: Optional[str] = None,
        freeze_n_epochs: int = 0,
    ):
        """

        Args:
            pairwise_model: Pairwise model to train
            pairs_miner: Miner of pairs
            criterion: Criterion to optimize
            optimizer: Optimizer
            scheduler: Learning rate scheduler
            scheduler_interval: Interval of calling scheduler (must be ``step`` or ``epoch``)
            scheduler_frequency: Frequency of calling scheduler
            input_tensors_key: Key to get tensors from the batches
            labels_key: Key to get labels from the batches
            embeddings_key: Key to get embeddings from the batches
            scheduler_monitor_metric: Metric to monitor for the schedulers that depend on the metric value
            freeze_n_epochs: number of epochs to freeze model (for n > 0 model has to be a successor of ``IFreezable``
                interface). When ``current_epoch >= freeze_n_epochs`` model is unfreezed. Note that epochs are
                starting with 0.

        """
        pl.LightningModule.__init__(self)

        self.model = pairwise_model
        self.miner = pairs_miner
        self.criterion = criterion
        self.optimizer = optimizer
        self.scheduler = scheduler
        self.scheduler_interval = scheduler_interval
        self.scheduler_frequency = scheduler_frequency
        self.input_tensors_key = input_tensors_key
        self.labels_key = labels_key
        self.embeddings_key = embeddings_key
        self.monitor_metric = scheduler_monitor_metric
        self.freeze_n_epochs = freeze_n_epochs

    def training_step(self, batch: Dict[str, Any], batch_idx: int) -> Tensor:
        ids1, ids2, is_negative = self.miner.sample(features=batch[self.embeddings_key], labels=batch[self.labels_key])
        x1 = batch[self.input_tensors_key][ids1]
        x2 = batch[self.input_tensors_key][ids2]
        target = is_negative.float()

        predictions = self.model(x1=x1, x2=x2)

        loss = self.criterion(predictions, target.to(predictions.device))

        bs = len(batch[self.labels_key])

        self.log("loss", loss.item(), prog_bar=True, batch_size=bs, on_step=True, on_epoch=True)

        if self.scheduler is not None:
            self.log("lr", self.scheduler.get_last_lr()[0], prog_bar=True, batch_size=bs, on_step=True, on_epoch=True)

        return loss

    def validation_step(self, batch: Dict[str, Any], batch_idx: int, *_: Any) -> Dict[str, Any]:
        # We simply accumulate batches here since we apply postprocessor during metrics calculation
        return batch

    def configure_optimizers(self) -> Any:
        if self.scheduler is None:
            return self.optimizer
        else:
            scheduler = {
                "scheduler": self.scheduler,
                "interval": self.scheduler_interval,
                "frequency": self.scheduler_frequency,
            }
            if isinstance(self.scheduler, ReduceLROnPlateau):
                scheduler["monitor"] = self.monitor_metric
            return [self.optimizer], [scheduler]

    def on_epoch_start(self) -> None:
        if self.freeze_n_epochs and isinstance(self.model, IFreezable):
            if self.current_epoch >= self.freeze_n_epochs:
                self.model.unfreeze()
            else:
                self.model.freeze()

    def get_progress_bar_dict(self) -> Dict[str, Union[int, str]]:
        # https://github.com/Lightning-AI/lightning/issues/1595
        tqdm_dict = super().get_progress_bar_dict()
        tqdm_dict.pop("v_num", None)
        return tqdm_dict


class PairwiseModuleDDP(PairwiseModule, ModuleDDP):
    def __init__(
        self,
        loaders_train: Optional[TRAIN_DATALOADERS] = None,
        loaders_val: Optional[EVAL_DATALOADERS] = None,
        *args: Any,
        **kwargs: Any,
    ):
        ModuleDDP.__init__(self, loaders_train=loaders_train, loaders_val=loaders_val)
        PairwiseModule.__init__(self, *args, **kwargs)


__all__ = ["PairwiseModule", "PairwiseModuleDDP"]



from typing import Any, Dict

from oml.const import TCfg
from oml.interfaces.retrieval import IRetrievalPostprocessor
from oml.registry.models import get_pairwise_model_by_cfg
from oml.retrieval.postprocessors.algo import AdaptiveThresholding, ConstantThresholding
from oml.retrieval.postprocessors.pairwise import PairwiseReranker
from oml.utils.misc import dictconfig_to_dict

POSTPROCESSORS_REGISTRY = {
    "pairwise_reranker": PairwiseReranker,
    "constant_thresholding": ConstantThresholding,
    "adaptive_thresholding": AdaptiveThresholding,
}


def get_postprocessor(name: str, **kwargs: Dict[str, Any]) -> IRetrievalPostprocessor:
    constructor = POSTPROCESSORS_REGISTRY[name]

    if "pairwise_model" in kwargs:
        kwargs["pairwise_model"] = get_pairwise_model_by_cfg(kwargs["pairwise_model"])  # type: ignore

    return constructor(**kwargs)  # type: ignore


def get_postprocessor_by_cfg(cfg: TCfg) -> IRetrievalPostprocessor:
    cfg = dictconfig_to_dict(cfg)
    postprocessor = get_postprocessor(cfg["name"], **cfg["args"])
    return postprocessor


__all__ = ["get_postprocessor", "get_postprocessor_by_cfg"]



from abc import abstractmethod

from matplotlib import pyplot as plt
from pytorch_lightning.loggers import Logger as LightningLogger

from oml.const import TCfg


class IFigureLogger:
    @abstractmethod
    def log_figure(self, fig: plt.Figure, title: str, idx: int) -> None:
        raise NotImplementedError()


class IPipelineLogger(LightningLogger, IFigureLogger):
    @abstractmethod
    def log_pipeline_info(self, cfg: TCfg) -> None:
        raise NotImplementedError()



from collections import Counter, defaultdict
from copy import deepcopy
from typing import Dict, Iterator, List, Set, Union

import numpy as np

from oml.interfaces.samplers import IBatchSampler
from oml.utils.misc import smart_sample


class DistinctCategoryBalanceSampler(IBatchSampler):
    """
    This sampler takes ``n_instances`` for each of the ``n_labels`` for each of the
    ``n_categories`` to form the batches.
    Thus, the batch size is ``n_instances x n_labels x n_categories``.

    The strategy for the dataset with ``L`` unique labels and ``C`` unique categories is the following:

    - Select ``n_categories`` of ``C`` for the 1st batch

    - Select ``n_labels`` for each of the chosen categories for the 1st batch

    - Select ``n_instances`` for each of the chosen labels for the 1st batch

    - Define the set of available for the 2nd batch labels ``L^``: these are all the labels ``L`` except the ones
      chosen for the 1st batch

    - Define set of available categories ``C^``: these are all the categories corresponding to labels from ``L^``

    - Select ``n_categories`` from ``C^`` for the 2nd batch

    - Select ``n_labels`` for each category from ``L^`` for the 2nd batch

    - Select ``n_instances`` for each label for the 2nd batch

    - ...

    - Epoch ends after ``epoch_size`` steps

    Behavior in corner cases:

    - If all the categories were chosen before ``epoch_size`` steps, the sampler resets its state and goes on sampling
     from the first step.

    - If some class does not contain ``n_instances``, a choice will be made with repetition.

    - If the chosen category does not contain unused ``n_labels``, all the unused labels will be added to a batch and
     the missing ones will be sampled from the used labels without repetition.

    - If ``L % n_labels == 1`` then one of the labels must be dropped because we always want to have more than 1 label
     in a batch to be able to form positive pairs later on.

    """

    def __init__(
        self,
        labels: Union[List[int], np.ndarray],
        label2category: Dict[int, Union[str, int]],
        n_categories: int,
        n_labels: int,
        n_instances: int,
        epoch_size: int,
    ):
        """

        Args:
            labels: Labels to sample from
            label2category: Mapping from label to category
            n_categories: The desired number of categories to sample for each batch
            n_labels: The desired number of labels to sample for each category in batch
            n_instances: The desired number of samples to sample for each label in batch
            epoch_size: The desired number of batches in epoch

        """
        unique_labels = set(labels)
        unique_categories = set(label2category.values())
        category2labels = {
            category: {label for label, cat in label2category.items() if category == cat}
            for category in sorted(list(unique_categories))
        }

        for param in [n_categories, n_labels, n_instances]:
            if not isinstance(param, int):
                raise TypeError(f"{param.__name__} must be int, {type(param)} given")
        if not 1 <= n_categories <= len(unique_categories):
            raise ValueError(f"must be 1 <= n_categories <= {len(unique_categories)}, {n_categories} given")
        if not 1 < n_labels <= len(unique_labels):
            raise ValueError(f"must be 1 < n_labels <= {len(unique_labels)}, {n_labels} given")
        if n_instances <= 1:
            raise ValueError(f"must be not less than 1, {n_instances} given")
        if any(label not in label2category.keys() for label in unique_labels):
            raise ValueError("All the labels must have category")
        if any(label not in unique_labels for label in label2category.keys()):
            raise ValueError("All the labels from label2category mapping must be in the labels")
        if any(n <= 1 for n in Counter(labels).values()):
            raise ValueError("Each class must contain at least 2 instances to fit")
        if any(len(list(labs)) < n_labels for labs in category2labels.values()):
            raise ValueError(f"All the categories must have at least {n_labels} unique labels")

        self._labels = np.array(labels)
        self._label2category = label2category
        self.n_categories = n_categories
        self.n_labels = n_labels
        self.n_instances = n_instances
        self._epoch_size = epoch_size

        self._batch_size = self.n_categories * self.n_labels * self.n_instances

        self._label2index = {
            label: np.arange(len(self._labels))[self._labels == label].tolist() for label in sorted(list(unique_labels))
        }
        self._category2labels = category2labels

    @property
    def batch_size(self) -> int:
        return self._batch_size

    def __len__(self) -> int:
        return self._epoch_size

    def __iter__(self) -> Iterator[List[int]]:
        category2labels = deepcopy(self._category2labels)
        used_labels: Dict[int, Set[int]] = defaultdict(set)
        epoch_indices = []
        for _ in range(self._epoch_size):
            if len(category2labels) < self.n_categories:
                category2labels = deepcopy(self._category2labels)
                used_labels = defaultdict(set)
            categories_available = list(category2labels.keys())
            categories = np.random.choice(
                categories_available, size=min(self.n_categories, len(categories_available)), replace=False
            )
            batch_indices = []
            for category in categories:
                labels_available = list(category2labels[category])
                labels_available_number = len(labels_available)
                if self.n_labels <= labels_available_number:
                    labels = np.random.choice(labels_available, size=self.n_labels, replace=False).tolist()
                else:
                    labels = (
                        labels_available
                        + np.random.choice(
                            list(used_labels[category]), size=self.n_labels - labels_available_number, replace=False
                        ).tolist()
                    )
                for label in labels:
                    indices = self._label2index[label]
                    samples_indices = smart_sample(array=indices, k=self.n_instances)
                    batch_indices.extend(samples_indices)
                category2labels[category] -= set(labels)
                used_labels[category].update(labels)
                if not category2labels[category]:
                    category2labels.pop(category)
            epoch_indices.append(batch_indices)
        return iter(epoch_indices)


__all__ = ["DistinctCategoryBalanceSampler"]



import ast
import importlib
import json
from itertools import chain
from pathlib import Path
from typing import List, Tuple

import pytest

from oml.const import PROJECT_ROOT

LIBS_TO_IGNORE = ["torch_xla", "pytorch_grad_cam", "wandb", "neptune", "clearml", "IPython", "transformers"]

NEED_TO_TEST_NOTEBOOKS = True


def get_imports_from_files() -> List[Tuple[str, str]]:
    files = get_files_with_imports()

    file_import_pairs = []

    for file in files:
        fname = PROJECT_ROOT / file
        if fname.suffix == ".py":
            imports = find_imports_in_script(fname)
        elif fname.suffix == ".ipynb":
            imports = find_imports_in_notebook(fname)
        else:
            raise ValueError

        file_import_pairs.extend(list(zip([file] * len(imports), imports)))

    return file_import_pairs


def get_files_with_imports() -> List[str]:
    files = []

    folder_with_scripts = PROJECT_ROOT / "oml"
    scriptes_files = sorted(str(fname.relative_to(PROJECT_ROOT)) for fname in folder_with_scripts.rglob("*.py"))
    files.extend(scriptes_files)

    folder_with_tests = PROJECT_ROOT / "tests"
    tests_files = sorted(str(fname.relative_to(PROJECT_ROOT)) for fname in folder_with_tests.rglob("*.py"))
    files.extend(tests_files)

    pipelines_folder = PROJECT_ROOT / "pipelines"
    pipelines_files = sorted(str(fname.relative_to(PROJECT_ROOT)) for fname in folder_with_tests.rglob("*.py"))
    files.extend(pipelines_files)

    notebooks_files = sorted(
        str(fname.relative_to(PROJECT_ROOT))
        for fname in pipelines_folder.rglob("*.ipynb")
        if fname.parent.name != ".ipynb_checkpoints"
    )

    if NEED_TO_TEST_NOTEBOOKS:
        files.extend(notebooks_files)

    return files


def find_imports_in_script(fname: Path) -> List[str]:
    with open(fname, "r") as f:
        script_code = f.read()

    imports = find_imports(script_code)

    return imports


def find_imports_in_notebook(fname: Path) -> List[str]:
    with open(fname, "r") as f:
        notebook_raw = json.load(f)

    code_lines = [cell["source"] for cell in notebook_raw["cells"] if cell["cell_type"] == "code"]
    code_lines = list(chain(*code_lines))
    # cells with single row haven't new line symbol
    code_lines = [line if line.endswith("\n") else f"{line}\n" for line in code_lines]
    # remove jupyter magic commands
    code_lines = filter(lambda x: not (x.startswith("%") or x.startswith("!")), code_lines)

    notebook_code = "".join(code_lines)
    imports = find_imports(notebook_code)

    return imports


def find_imports(code: str) -> List[str]:
    code = ast.parse(code)
    imports = set()
    for node in ast.walk(code):
        if isinstance(node, ast.Import):
            for name in node.names:
                imports.add(name.name)
        elif isinstance(node, ast.ImportFrom) and node.level == 0:
            imports.add(node.module)
    return list(imports)


@pytest.mark.parametrize("file,lib", get_imports_from_files())
def test_project_imports(file: str, lib: str) -> None:
    if any(lib.startswith(ignore_lib) for ignore_lib in LIBS_TO_IGNORE):
        pass
    else:
        importlib.import_module(lib)



import ast
from pathlib import Path
from typing import List, Optional, Tuple

from oml.const import PROJECT_ROOT


def parse_arg_in_docstring_row(row: str) -> Tuple[Optional[str], Optional[int]]:
    # We assume Google docstring format
    if ":" not in row:
        return None, None
    else:
        prefix = row.split(":")[0]
        offset = len(prefix) - len(prefix.lstrip(" "))
        return prefix[offset:], offset


def parse_args_in_docstring(docstring: str) -> List[str]:
    # 1. Find first docstring row if exists
    row_first_arg = None
    docstring_rows = docstring.split("\n")

    for i, row in enumerate(docstring_rows):
        if "Args:" in row:
            row_first_arg = i + 1
            break

    parsed_args: List[str] = []

    if row_first_arg is None:
        return parsed_args

    # 2. Parse the row that has to contain the first argument and check its offset
    first_arg, first_offset = parse_arg_in_docstring_row(docstring_rows[row_first_arg])
    assert first_arg
    parsed_args.append(first_arg)

    # 3. Parse rest of the rows, check the corresponding offsets
    for row in docstring_rows[row_first_arg + 1 :]:
        if ("Returns:" in row) or ("Raises:" in row):
            break

        arg, offset = parse_arg_in_docstring_row(row)
        if arg and (offset == first_offset):
            parsed_args.append(arg)

    return parsed_args


def check_docstrings_in_file(filename: Path) -> None:
    with open(filename, "r") as file:
        content = file.read()

    tree = ast.parse(content)

    for node in ast.walk(tree):
        if isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef)):
            docstring = ast.get_docstring(node)

            if docstring is not None:
                actual_args = set([arg.arg for arg in node.args.args]) - {"self", "cls"}
                docstring_args = set(parse_args_in_docstring(docstring))

                if docstring_args and (actual_args != docstring_args) and ("*_" not in docstring_args):
                    raise ValueError(
                        f"Incorrect docstring for {node.name} in {filename}."
                        f"Actual args are: {actual_args}\n"
                        f"Docstring args are: {docstring_args}"
                    )


def check_docstrings_in_dir(directory: Path) -> None:
    for fname in Path(directory).glob("**/*.py"):
        check_docstrings_in_file(fname)


def test_docstrings() -> None:
    check_docstrings_in_dir(PROJECT_ROOT / "oml")



from typing import Optional, Sequence, Tuple

import torch
from torch import BoolTensor, FloatTensor, LongTensor, Tensor
from tqdm.auto import tqdm

from oml.const import BS_KNN
from oml.utils.misc_torch import pairwise_dist


def check_both_are_nones_or_not_nones(a: Optional[Tensor], b: Optional[Tensor]) -> bool:
    return (a is not None and b is not None) or (a is None and b is None)


def check_len_and_ndim(a: Optional[Tensor], expected_len: int, expected_ndim: int) -> bool:
    return (a is None) or (len(a) == expected_len and a.ndim == expected_ndim)


def batched_knn_qg(
    embeddings_query: FloatTensor,
    embeddings_gallery: FloatTensor,
    top_n: int,
    labels_query: Optional[LongTensor] = None,
    labels_gallery: Optional[LongTensor] = None,
    ids_query: Optional[LongTensor] = None,
    ids_gallery: Optional[LongTensor] = None,
    sequence_ids_query: Optional[LongTensor] = None,
    sequence_ids_gallery: Optional[LongTensor] = None,
    bs: int = BS_KNN,
    verbose: bool = False,
) -> Tuple[Sequence[FloatTensor], Sequence[LongTensor], Optional[Sequence[LongTensor]]]:
    """
    Arguments have the same meaning as in `batched_knn`.

    Note, since queries and galleries are separated we additionally need
    `ids_query` and `ids_gallery` in order to process the situation when the same items
    are in the query and are in the gallery. If `ids_query` and `ids_gallery` don't overlap,
    just set both of them to Nones.

    """
    nq = len(embeddings_query)
    ng = len(embeddings_gallery)

    assert (embeddings_query.ndim == 2) and (embeddings_gallery.ndim == 2)
    assert check_both_are_nones_or_not_nones(labels_query, labels_gallery)
    assert check_both_are_nones_or_not_nones(ids_query, ids_gallery)
    assert check_both_are_nones_or_not_nones(sequence_ids_query, sequence_ids_gallery)

    assert check_len_and_ndim(labels_query, nq, 1)
    assert check_len_and_ndim(labels_gallery, ng, 1)
    assert check_len_and_ndim(ids_query, nq, 1)
    assert check_len_and_ndim(ids_gallery, ng, 1)
    assert check_len_and_ndim(sequence_ids_query, nq, 1)
    assert check_len_and_ndim(sequence_ids_gallery, ng, 1)

    top_n = min(top_n, ng)

    retrieved_ids = []
    distances = []
    gt_ids = []

    # we do batching over first (queries) dimension
    items = tqdm(range(0, nq, bs), desc="Finding nearest neighbors.") if verbose else range(0, nq, bs)
    for i in items:
        distances_b = pairwise_dist(x1=embeddings_query[i : i + bs, :], x2=embeddings_gallery, p=2)

        mask_to_ignore_b = torch.zeros_like(distances_b, dtype=torch.bool).to(distances_b.device)

        if (ids_query is not None) and (ids_gallery is not None):
            # we want to ignore the item during search if it was used for both: query and gallery
            mask_to_ignore_same_item = ids_query[i : i + bs][..., None] == ids_gallery[None, ...]
            mask_to_ignore_b = torch.logical_or(mask_to_ignore_b, mask_to_ignore_same_item)

        if (sequence_ids_query is not None) and (sequence_ids_gallery is not None):
            # our items may be packed into the sequences, so we ignore other members of this sequence during search
            # more info in the docs: search for "Handling sequences of photos"
            mask_sequence = sequence_ids_query[i : i + bs][..., None] == sequence_ids_gallery[None, ...]
            mask_to_ignore_b = torch.logical_or(mask_to_ignore_b, mask_sequence)

        if (labels_query is not None) and (labels_gallery is not None):
            mask_gt_b = BoolTensor(labels_query[i : i + bs][..., None] == labels_gallery[None, ...])
            mask_gt_b[mask_to_ignore_b] = False
            gt_ids.extend([row.nonzero().view(-1) for row in mask_gt_b])  # type: ignore

        distances_b[mask_to_ignore_b] = float("inf")
        distances_b_sorted, retrieved_ids_b = torch.topk(distances_b, k=top_n, largest=False, sorted=True)

        # every query may have arbitrary number of retrieved items, so we are forced to use a loop to store the results
        for dist, ids in zip(distances_b_sorted, retrieved_ids_b):
            mask_to_keep = ~dist.isinf()
            distances.append(dist[mask_to_keep].view(-1))
            retrieved_ids.append(ids[mask_to_keep].view(-1))

    return distances, retrieved_ids, gt_ids or None


def batched_knn(
    embeddings: FloatTensor,
    ids_query: LongTensor,
    ids_gallery: LongTensor,
    top_n: int,
    sequence_ids: Optional[LongTensor] = None,
    labels_gt: Optional[LongTensor] = None,
    bs: int = BS_KNN,
    verbose: bool = False,
) -> Tuple[Sequence[FloatTensor], Sequence[LongTensor], Optional[Sequence[LongTensor]]]:
    """

    Args:
        embeddings: Matrix with the shape of ``[L, dim]``
        ids_query:  Tensor with the size of ``Q``, where ``Q <= n``. Each element is within the range ``(0, L - 1)``.
        ids_gallery:  Tensor with the size of ``G`` where ``G <= n``. Each element is within the range ``(0, L - 1)``.
                      May overlap with ``ids_query``.
        top_n: Number of neighbors to retrieve.
        sequence_ids: Sequence identifiers with the size of ``L`` (if known).
        labels_gt: Ground truth labels of every element with the size of ``L`` (if known).
        bs: Batch size for computing distances to avoid OOM errors when processing the whole matrix at once.
        verbose: Set ``True`` to see progress bar.

    Returns:
        distances: Sorted distances from queries to the first gallery items with the size of ``Q``.
        retrieved_ids: First gallery indices retrieved for every query with the size of ``Q``.
                Every index is within the range ``(0, G - 1)``.
        gt_ids: Gallery indices relevant to every query with the size of ``Q``.
                Every element is within the range ``(0, G - 1)``

    """
    assert (ids_query.ndim == 1) and (ids_gallery.ndim == 1) and (embeddings.ndim == 2)
    assert len(embeddings) <= len(ids_query) + len(ids_gallery)
    assert check_len_and_ndim(sequence_ids, len(embeddings), 1)
    assert check_len_and_ndim(labels_gt, len(embeddings), 1)

    return batched_knn_qg(
        embeddings_query=embeddings[ids_query],
        embeddings_gallery=embeddings[ids_gallery],
        ids_query=ids_query,
        ids_gallery=ids_gallery,
        top_n=top_n,
        bs=bs,
        verbose=verbose,
        labels_query=None if labels_gt is None else labels_gt[ids_query],
        labels_gallery=None if labels_gt is None else labels_gt[ids_gallery],
        sequence_ids_query=None if sequence_ids is None else sequence_ids[ids_query],
        sequence_ids_gallery=None if sequence_ids is None else sequence_ids[ids_gallery],
    )


__all__ = ["batched_knn", "batched_knn_qg"]



from pathlib import Path
from typing import Optional, Union

import numpy as np
import PIL
import torch
from PIL.Image import Image as TPILImage
from torch import nn

from oml.const import MEAN, STD, STORAGE_CKPTS, TNormParam
from oml.interfaces.models import IExtractor
from oml.models.utils import (
    remove_criterion_in_state_dict,
    remove_prefix_from_state_dict,
)
from oml.models.vit_dino.external.hubconf import (  # type: ignore
    dino_vitb8,
    dino_vitb16,
    dino_vits8,
    dino_vits16,
)
from oml.models.vit_dino.external_v2.hubconf import (  # type: ignore
    dinov2_vitb14,
    dinov2_vitb14_reg,
    dinov2_vitl14,
    dinov2_vitl14_reg,
    dinov2_vits14,
    dinov2_vits14_reg,
)
from oml.transforms.images.albumentations import get_normalisation_albu
from oml.utils.io import download_checkpoint_one_of
from oml.utils.misc_torch import normalise, temporary_setting_model_mode

_FB_URL = "https://dl.fbaipublicfiles.com"


class ViTExtractor(IExtractor):
    """
    The base class for the extractors that follow VisualTransformer architecture.

    """

    constructors = {
        "vits8": dino_vits8,
        "vits16": dino_vits16,
        "vitb8": dino_vitb8,
        "vitb16": dino_vitb16,
        "vits14": dinov2_vits14,
        "vitb14": dinov2_vitb14,
        "vitl14": dinov2_vitl14,
        "vits14_reg": dinov2_vits14_reg,
        "vitb14_reg": dinov2_vitb14_reg,
        "vitl14_reg": dinov2_vitl14_reg,
    }

    pretrained_models = {
        # checkpoints pretrained in DINO framework on ImageNet by MetaAI
        "vits16_dino": {
            "url": f"{_FB_URL}/dino/dino_deitsmall16_pretrain/dino_deitsmall16_pretrain.pth",
            "hash": "cf0f22",
            "fname": "vits16_dino.ckpt",
            "init_args": {"arch": "vits16", "normalise_features": False},
        },
        "vits8_dino": {
            "url": f"{_FB_URL}/dino/dino_deitsmall8_pretrain/dino_deitsmall8_pretrain.pth",
            "hash": "230cd5",
            "fname": "vits8_dino.ckpt",
            "init_args": {"arch": "vits8", "normalise_features": False},
        },
        "vitb16_dino": {
            "url": f"{_FB_URL}/dino/dino_vitbase16_pretrain/dino_vitbase16_pretrain.pth",
            "hash": "552daf",
            "fname": "vitb16_dino.ckpt",
            "init_args": {"arch": "vitb16", "normalise_features": False},
        },
        "vitb8_dino": {
            "url": f"{_FB_URL}/dino/dino_vitbase8_pretrain/dino_vitbase8_pretrain.pth",
            "hash": "556550",
            "fname": "vitb8_dino.ckpt",
            "init_args": {"arch": "vitb8", "normalise_features": False},
        },
        "vits14_dinov2": {
            "url": f"{_FB_URL}/dinov2/dinov2_vits14/dinov2_vits14_pretrain.pth",
            "hash": "2e405c",
            "fname": "dinov2_vits14.ckpt",
            "init_args": {"arch": "vits14", "normalise_features": False},
        },
        "vits14_reg_dinov2": {
            "url": f"{_FB_URL}/dinov2/dinov2_vits14/dinov2_vits14_reg4_pretrain.pth",
            "hash": "2a50c5",
            "fname": "dinov2_vits14_reg4.ckpt",
            "init_args": {"arch": "vits14_reg", "normalise_features": False},
        },
        "vitb14_dinov2": {
            "url": f"{_FB_URL}/dinov2/dinov2_vitb14/dinov2_vitb14_pretrain.pth",
            "hash": "8635e7",
            "fname": "dinov2_vitb14.ckpt",
            "init_args": {"arch": "vitb14", "normalise_features": False},
        },
        "vitb14_reg_dinov2": {
            "url": f"{_FB_URL}/dinov2/dinov2_vitb14/dinov2_vitb14_reg4_pretrain.pth",
            "hash": "13d13c",
            "fname": "dinov2_vitb14_reg4.ckpt",
            "init_args": {"arch": "vitb14_reg", "normalise_features": False},
        },
        "vitl14_dinov2": {
            "url": f"{_FB_URL}/dinov2/dinov2_vitl14/dinov2_vitl14_pretrain.pth",
            "hash": "19a02c",
            "fname": "dinov2_vitl14.ckpt",
            "init_args": {"arch": "vitl14", "normalise_features": False},
        },
        "vitl14_reg_dinov2": {
            "url": f"{_FB_URL}/dinov2/dinov2_vitl14/dinov2_vitl14_reg4_pretrain.pth",
            "hash": "8b6364",
            "fname": "dinov2_vitl14_reg4.ckpt",
            "init_args": {"arch": "vitl14_reg", "normalise_features": False},
        },
        # our pretrained checkpoints
        "vits16_inshop": {
            "url": [
                f"{STORAGE_CKPTS}/inshop/vits16_inshop_a76b85.ckpt",
                "1niX-TC8cj6j369t7iU2baHQSVN3MVJbW",
            ],
            "hash": "a76b85",
            "fname": "vits16_inshop.ckpt",
            "init_args": {"arch": "vits16", "normalise_features": False},
        },
        "vits16_sop": {
            "url": [
                f"{STORAGE_CKPTS}/sop/vits16_sop_21e743.ckpt",
                "1zuGRHvF2KHd59aw7i7367OH_tQNOGz7A",
            ],
            "hash": "21e743",
            "fname": "vits16_sop.ckpt",
            "init_args": {"arch": "vits16", "normalise_features": True},
        },
        "vits16_cub": {
            "url": [
                f"{STORAGE_CKPTS}/cub/vits16_cub.ckpt",
                "1p2tUosFpGXh5sCCdzlXtjV87kCDfG34G",
            ],
            "hash": "e82633",
            "fname": "vits16_cub.ckpt",
            "init_args": {"arch": "vits16", "normalise_features": False},
        },
        "vits16_cars": {
            "url": [
                f"{STORAGE_CKPTS}/cars/vits16_cars.ckpt",
                "1hcOxDRRXrKr6ZTCyBauaY8Ue-pok4Icg",
            ],
            "hash": "9f1e59",
            "fname": "vits16_cars.ckpt",
            "init_args": {"arch": "vits16", "normalise_features": False},
        },
    }

    def __init__(
        self,
        weights: Optional[Union[Path, str]],
        arch: str,
        normalise_features: bool,
        use_multi_scale: bool = False,
    ):
        """
        Args:
            weights: Path to weights or a special key to download pretrained checkpoint, use ``None`` to
             randomly initialize model's weights. You can check the available pretrained checkpoints
             in ``self.pretrained_models``.
            arch: Might be one of ``vits8``, ``vits16``, ``vitb8``, ``vitb16``. You can check all the available options
             in ``self.constructors``
            normalise_features: Set ``True`` to normalise output features
            use_multi_scale: Set ``True`` to use multiscale (the analogue of test time augmentations)

        """
        assert arch in self.constructors
        super(ViTExtractor, self).__init__()

        self.normalise_features = normalise_features
        self.mscale = use_multi_scale
        self.arch = arch

        factory_fun = self.constructors[self.arch]

        self.model = factory_fun(pretrained=False)
        if weights is None:
            return

        if weights in self.pretrained_models:
            pretrained = self.pretrained_models[weights]  # type: ignore
            weights = download_checkpoint_one_of(
                url_or_fid_list=pretrained["url"],  # type: ignore
                hash_md5=pretrained["hash"],  # type: ignore
                fname=pretrained["fname"],  # type: ignore
            )

        ckpt = torch.load(weights, map_location="cpu")
        state_dict = ckpt["state_dict"] if "state_dict" in ckpt else ckpt
        state_dict = remove_criterion_in_state_dict(state_dict)
        ckpt = remove_prefix_from_state_dict(state_dict, trial_key="norm.bias")
        self.model.load_state_dict(ckpt, strict=True)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if self.mscale:
            x = self._multi_scale(x)
        else:
            x = self.model(x)

        if self.normalise_features:
            x = normalise(x)

        return x

    @property
    def feat_dim(self) -> int:
        return len(self.model.norm.bias)

    def _multi_scale(self, samples: torch.Tensor) -> torch.Tensor:
        # code from the original DINO
        # TODO: check grads later
        v = torch.zeros((len(samples), self.feat_dim), device=samples.device)
        scales = [1.0, 1 / 2 ** (1 / 2), 1 / 2]  # we use 3 different scales
        for s in scales:
            if s == 1:
                inp = samples.clone()
            else:
                inp = nn.functional.interpolate(samples, scale_factor=s, mode="bilinear", align_corners=False)
            feats = self.model.forward(inp).clone()
            v += feats

        v /= len(scales)
        # v /= v.norm(dim=1)  # we don't want to shift the norms values
        return v

    def draw_attention(self, image: Union[TPILImage, np.ndarray]) -> np.ndarray:
        """
        Args:
            image: An image with pixel values in the range of ``[0..255]``.

        Returns:
            An image with drawn attention maps.

        Visualization of the multi-head attention on a particular image.

        """
        return vis_vit(vit=self, image=image)


def vis_vit(
    vit: ViTExtractor,
    image: Union[TPILImage, np.ndarray],
    mean: TNormParam = MEAN,
    std: TNormParam = STD,
) -> np.ndarray:
    # this is the optional dependency
    from pytorch_grad_cam.utils.image import show_cam_on_image

    need_to_convert = not isinstance(image, np.ndarray)

    if need_to_convert:
        image = np.asarray(image)

    patch_size = vit.model.patch_embed.proj.kernel_size[0]

    img_tensor = get_normalisation_albu(mean=mean, std=std)(image=image)["image"]

    w = img_tensor.shape[1] - img_tensor.shape[1] % patch_size
    h = img_tensor.shape[2] - img_tensor.shape[2] % patch_size

    img_tensor = img_tensor[:, :w, :h].unsqueeze(0)

    w_feat_map = img_tensor.shape[-2] // patch_size
    h_feat_map = img_tensor.shape[-1] // patch_size

    with temporary_setting_model_mode(vit, set_train=False):
        with torch.no_grad():
            attentions = vit.model.get_last_selfattention(img_tensor)

    nh = attentions.shape[1]

    attentions = attentions[0, :, 0, 1:].reshape(nh, -1)

    attentions = attentions.reshape(nh, w_feat_map, h_feat_map)
    attentions = (
        nn.functional.interpolate(
            attentions.unsqueeze(0),
            scale_factor=patch_size,
            mode="nearest",
        )[0]
        .cpu()
        .numpy()
    )

    arr = sum(attentions[i] * 1 / attentions.shape[0] for i in range(attentions.shape[0]))

    arr = show_cam_on_image(image / image.max(), 0.6 * arr / arr.max())  # type: ignore

    if need_to_convert:
        arr = PIL.Image.fromarray(arr)

    return arr


__all__ = ["ViTExtractor", "vis_vit"]



from oml.miners.cross_batch import TripletMinerWithMemory
from oml.miners.inbatch_all_tri import AllTripletsMiner, get_available_triplets
from oml.miners.inbatch_hard_cluster import HardClusterMiner
from oml.miners.inbatch_hard_tri import HardTripletsMiner
from oml.miners.inbatch_nhard_tri import NHardTripletsMiner
from oml.miners.miner_with_bank import MinerWithBank
from oml.miners.pairs import PairsMiner



from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
import torch
from torch import Tensor

from oml.ddp.utils import get_world_size_safe, sync_dicts_ddp
from oml.utils.misc_torch import unique_by_ids

TStorage = Dict[str, Union[Tensor, np.ndarray, List[Any]]]


class Accumulator:
    def __init__(self, keys_to_accumulate: Tuple[str, ...]):
        """
        Class for accumulating values of different types, for instance,
        torch.Tensor and numpy.array.

        Args:
            keys_to_accumulate: List or tuple of keys to be collected.
                                 We will take values via these keys calling ``self.update_data()``.
        """
        assert len(keys_to_accumulate) == len(set(keys_to_accumulate)), "All the keys have to be unique!"

        self.keys_to_accumulate = keys_to_accumulate
        self.num_samples = None

        self._collected_samples = 0
        self._storage: TStorage = dict()

        self._indices_key = "__element_indices"  # internal key to keep track of elements order if provided

    def refresh(self, num_samples: int) -> None:
        """
        This method refreshes the state.

        Args:
            num_samples:  The total number of elements you are going to collect (for memory allocation).

        """
        assert isinstance(num_samples, int) and num_samples > 0
        self.num_samples = num_samples  # type: ignore
        self._collected_samples = 0
        self._storage = {}

    def _allocate_memory_if_need(self, key: str, batch_value: Any) -> None:
        if self.num_samples is None:
            raise ValueError(
                f"The parameter for memory allocation has not been set up."
                f"Are you sure you've called {self.refresh.__name__}?"
            )

        if key not in self._storage:
            if isinstance(batch_value, torch.Tensor):
                self._storage[key] = torch.empty(
                    (self.num_samples, *batch_value.shape[1:]),
                    dtype=batch_value.dtype,
                    device="cpu",
                    requires_grad=False,
                )
            elif isinstance(batch_value, np.ndarray):
                self._storage[key] = np.empty((self.num_samples, *batch_value.shape[1:]), dtype=batch_value.dtype)
            elif isinstance(batch_value, (list, tuple)):
                self._storage[key] = []
            else:
                raise TypeError(f"Type '{type(batch_value)}' is not available for accumulating")

    def _put_in_storage(self, key: str, batch_value: Any) -> None:
        bs = len(batch_value)

        if isinstance(batch_value, torch.Tensor):
            bv = batch_value.detach()
            self._storage[key][self._collected_samples : self._collected_samples + bs, ...] = bv  # type: ignore
        elif isinstance(batch_value, np.ndarray):
            bv = batch_value
            self._storage[key][self._collected_samples : self._collected_samples + bs, ...] = bv  # type: ignore
        elif isinstance(batch_value, (list, tuple)):
            self._storage[key].extend(list(batch_value))  # type: ignore
        else:
            raise TypeError(f"Type '{type(batch_value)}' is not available for accumulating")

    def update_data(self, data_dict: Dict[str, Any], indices: Optional[List[int]] = None) -> None:
        """
        Args:
            data_dict: We will accumulate data getting values via ``self.keys_to_accumulate``. All elements
                       of the dictionary have to have the same size.
            indices: Global indices of the elements in your batch of data. If provided, the accumulator
                     will remove accumulated duplicates and return the elements in the sorted order after ``.sync()``.
                     Indices may be useful in DDP (because data is gathered shuffled, additionally you may also get
                     some duplicates due to padding). In the single device regime it's also useful if you accumulate
                     data in shuffled order.

        """
        keys = list(self.keys_to_accumulate)

        if indices is None:
            assert self._indices_key not in self.storage, "We are tracking ids, but they are not currently provided."
        else:
            assert isinstance(indices, List)
            if (self.collected_samples > 0) and (self._indices_key not in self.storage):
                raise RuntimeError("You provided ids, but seems like you had not done it before.")

            keys += [self._indices_key]
            data_dict[self._indices_key] = indices

        bs_values = [len(data_dict[k]) for k in keys]
        bs = bs_values[0]
        assert all(bs == bs_value for bs_value in bs_values), f"Lengths of data are not equal, lengths: {bs_values}"

        for k in keys:
            v = data_dict[k]
            self._allocate_memory_if_need(k, v)
            self._put_in_storage(k, v)

        self._collected_samples += bs

    @property
    def storage(self) -> TStorage:
        return self._storage

    @property
    def collected_samples(self) -> int:
        return self._collected_samples

    def is_storage_full(self) -> bool:
        return self.num_samples == self.collected_samples

    def sync(self) -> "Accumulator":
        """
        The method drops duplicates and sort elements by indices if they have been provided in ``self.update_data()``.
        In DDP it also gathers data collected on several devices.

        """
        # TODO: add option to broadcast instead of sync to avoid duplicating data
        if not self.is_storage_full():
            raise ValueError(f"Cannot sync. Collected: {self.num_samples}/{self.collected_samples} items.")

        params = {"num_samples": [self.num_samples], "keys_to_accumulate": self.keys_to_accumulate}
        storage = self._storage

        world_size = get_world_size_safe()
        need_rebuilding = False

        if world_size > 1:
            params = sync_dicts_ddp(params, world_size=world_size, device="cpu")
            storage = sync_dicts_ddp(self._storage, world_size=world_size, device="cpu")
            need_rebuilding = True

            assert set(params["keys_to_accumulate"]) == set(
                self.keys_to_accumulate
            ), "Keys of accumulators should be the same on each device"

        if self._indices_key in storage:
            for key, data in storage.items():
                storage[key] = unique_by_ids(storage[self._indices_key], data)[1]  # type: ignore
            indices = storage[self._indices_key]
            need_rebuilding = True
        else:
            indices = None

        if not need_rebuilding:
            # If indices were not provided & it's not DDP we may save time & memory avoiding re-building accumulator
            return self

        synced_accum = Accumulator(tuple(set(params["keys_to_accumulate"])))
        synced_accum.refresh(num_samples=len(storage[list(storage.keys())[0]]))
        synced_accum.update_data(storage, indices=indices)

        return synced_accum


__all__ = ["TStorage", "Accumulator"]



from oml.losses.arcface import ArcFaceLoss, ArcFaceLossWithMLP
from oml.losses.surrogate_precision import SurrogatePrecision
from oml.losses.triplet import TripletLoss, TripletLossPlain, TripletLossWithMiner



# type: ignore
# flake8: noqa


from enum import Enum
from typing import Union

import torch

import oml.models.vit_dino.external_v2.vision_transformer as vits
from oml.const import CKPT_SAVE_ROOT

# ============== CODE FROM DINOV2 ==============
# https://github.com/facebookresearch/dinov2/blob/main/dinov2/hub/backbones.py

_DINOV2_BASE_URL = "https://dl.fbaipublicfiles.com/dinov2"


def _make_dinov2_model_name(arch_name: str, patch_size: int, num_register_tokens: int = 0) -> str:
    compact_arch_name = arch_name.replace("_", "")[:4]
    registers_suffix = f"_reg{num_register_tokens}" if num_register_tokens else ""
    return f"dinov2_{compact_arch_name}{patch_size}{registers_suffix}"


class Weights(Enum):
    LVD142M = "LVD142M"


def _make_dinov2_model(
    *,
    arch_name: str = "vit_large",
    img_size: int = 518,
    patch_size: int = 14,
    init_values: float = 1.0,
    ffn_layer: str = "mlp",
    block_chunks: int = 0,
    num_register_tokens: int = 0,
    interpolate_antialias: bool = False,
    interpolate_offset: float = 0.1,
    pretrained: bool = True,
    weights: Union[Weights, str] = Weights.LVD142M,
    **kwargs,
):
    if isinstance(weights, str):
        try:
            weights = Weights[weights]
        except KeyError as e:
            raise AssertionError(f"Unsupported weights: {weights}") from e

    model_base_name = _make_dinov2_model_name(arch_name, patch_size)
    vit_kwargs = {
        "img_size": img_size,
        "patch_size": patch_size,
        "init_values": init_values,
        "ffn_layer": ffn_layer,
        "block_chunks": block_chunks,
        "num_register_tokens": num_register_tokens,
        "interpolate_antialias": interpolate_antialias,
        "interpolate_offset": interpolate_offset,
    }
    vit_kwargs.update(**kwargs)
    model = vits.__dict__[arch_name](**vit_kwargs)

    if pretrained:
        model_full_name = _make_dinov2_model_name(arch_name, patch_size, num_register_tokens)
        url = _DINOV2_BASE_URL + f"/{model_base_name}/{model_full_name}_pretrain.pth"
        state_dict = torch.hub.load_state_dict_from_url(
            url, map_location="cpu", model_dir=str(CKPT_SAVE_ROOT.resolve())
        )
        model.load_state_dict(state_dict, strict=True)

    return model


def dinov2_vits14(*, pretrained: bool = True, weights: Union[Weights, str] = Weights.LVD142M, **kwargs):
    """
    DINOv2 ViT-S/14 model (optionally) pretrained on the LVD-142M dataset.
    """
    return _make_dinov2_model(arch_name="vit_small", pretrained=pretrained, weights=weights, **kwargs)


def dinov2_vitb14(*, pretrained: bool = True, weights: Union[Weights, str] = Weights.LVD142M, **kwargs):
    """
    DINOv2 ViT-B/14 model (optionally) pretrained on the LVD-142M dataset.
    """
    return _make_dinov2_model(arch_name="vit_base", pretrained=pretrained, weights=weights, **kwargs)


def dinov2_vitl14(*, pretrained: bool = True, weights: Union[Weights, str] = Weights.LVD142M, **kwargs):
    """
    DINOv2 ViT-L/14 model (optionally) pretrained on the LVD-142M dataset.
    """
    return _make_dinov2_model(arch_name="vit_large", pretrained=pretrained, weights=weights, **kwargs)


def dinov2_vits14_reg(*, pretrained: bool = True, weights: Union[Weights, str] = Weights.LVD142M, **kwargs):
    """
    DINOv2 ViT-S/14 model with registers (optionally) pretrained on the LVD-142M dataset
    """
    return _make_dinov2_model(
        arch_name="vit_small",
        pretrained=pretrained,
        weights=weights,
        num_register_tokens=4,
        interpolate_antialias=True,
        interpolate_offset=0.0,
        **kwargs,
    )


def dinov2_vitb14_reg(*, pretrained: bool = True, weights: Union[Weights, str] = Weights.LVD142M, **kwargs):
    """
    DINOv2 ViT-B/14 model with registers (optionally) pretrained on the LVD-142M dataset
    """
    return _make_dinov2_model(
        arch_name="vit_base",
        pretrained=pretrained,
        weights=weights,
        num_register_tokens=4,
        interpolate_antialias=True,
        interpolate_offset=0.0,
        **kwargs,
    )


def dinov2_vitl14_reg(*, pretrained: bool = True, weights: Union[Weights, str] = Weights.LVD142M, **kwargs):
    """
    DINOv2 ViT-L/14 model with registers (optionally) pretrained on the LVD-142M dataset
    """
    return _make_dinov2_model(
        arch_name="vit_large",
        pretrained=pretrained,
        weights=weights,
        num_register_tokens=4,
        interpolate_antialias=True,
        interpolate_offset=0.0,
        **kwargs,
    )



from functools import partial
from typing import List, Tuple, Union

import cv2
import numpy as np
import torch

SizeHW = Tuple[int, int]  # Size of images in (height, width) order
BboxesType = Union[torch.Tensor, np.ndarray]  # in format Nx4: [[left_1, top_1, right_1, bottom_1], ...]
ImageType = Union[torch.Tensor, np.ndarray]  # ndarray shapes: HxW, HxWxC. tensor shapes : HxW, CxHxW, BxCxHxW


def karesize_image(image: ImageType, new_hw: SizeHW) -> ImageType:
    """
    Function resizes image with keeping aspect ratio

    NOTE: tensors are resized on same device with gradient calculation

    """
    if isinstance(image, torch.Tensor):
        is_tensor = True
    elif isinstance(image, np.ndarray):
        is_tensor = False
    else:
        raise TypeError("Unsupported type")

    src_hw = get_image_hw(image)

    if new_hw == src_hw:
        return image

    _, resize_hw, padding = get_karesize_param(src_hw, new_hw)

    if is_tensor:
        image = resize_tensor(image, resize_hw)
        image = torch.nn.functional.pad(image, padding, mode="constant")
    else:
        pad_left, pad_right, pad_top, pad_bottom = padding
        padding_np = ((pad_top, pad_bottom), (pad_left, pad_right))
        if image.ndim == 3:
            padding_np = (*padding_np, (0, 0))  # type: ignore
        image = resize_ndarray(image, resize_hw)
        image = np.pad(image, padding_np)

    return image


def inverse_karesize_image(image: ImageType, src_hw: SizeHW) -> ImageType:
    """
    Function resizes image back to src size. The function inverse to the 'karesize_image' function

    NOTE: tensors are resized on same device with gradient calculation

    """
    if isinstance(image, torch.Tensor):
        is_tensor = True
    elif isinstance(image, np.ndarray):
        is_tensor = False
    else:
        raise TypeError("Unsupported type")

    curr_hw = get_image_hw(image)

    if curr_hw == src_hw:
        return image

    _, (resized_h, resized_w), (pad_left, _, pad_top, _) = get_karesize_param(src_hw, curr_hw)

    if is_tensor:
        image = image[..., pad_top : pad_top + resized_h, pad_left : pad_left + resized_w]
        image = resize_tensor(image, src_hw)
    else:
        image = image[pad_top : pad_top + resized_h, pad_left : pad_left + resized_w, ...]
        image = resize_ndarray(image, src_hw)

    return image


def karesize_bboxes(src_bboxes: BboxesType, src_hw: SizeHW, new_hw: SizeHW) -> BboxesType:
    """
    Function resizes bboxes with keeping aspect ratio from src size to desired size

    NOTE: tensors are resized on same device with gradient calculation

    """
    _assert_bbox_type_and_shape(src_bboxes)
    scale, _, (pad_left, _, pad_top, _) = get_karesize_param(src_hw, new_hw)

    bboxes = src_bboxes * scale
    bboxes[:, 0] += pad_left
    bboxes[:, 2] += pad_left
    bboxes[:, 1] += pad_top
    bboxes[:, 3] += pad_top

    return bboxes


def inverse_karesize_bboxes(cur_bboxes: BboxesType, curr_hw: SizeHW, src_hw: SizeHW) -> BboxesType:
    """
    Function resizes bboxes back to src size. The function inverse to the 'karesize_bboxes' function

    NOTE: tensors are resized on same device with gradient calculation

    """
    _assert_bbox_type_and_shape(cur_bboxes)
    scale, _, (pad_left, _, pad_top, _) = get_karesize_param(src_hw, curr_hw)

    padding = [[pad_left, pad_top, pad_left, pad_top]]

    if isinstance(cur_bboxes, np.ndarray):
        constructor = np.array
    elif isinstance(cur_bboxes, torch.Tensor):
        constructor = partial(torch.tensor, device=cur_bboxes.device)
    else:
        raise TypeError("Unsuported type of bboxes")

    padding = constructor(padding)

    src_bboxes = (cur_bboxes - padding) / scale

    return src_bboxes


def _assert_bbox_type_and_shape(cur_bboxes: BboxesType) -> None:
    assert isinstance(cur_bboxes, (torch.Tensor, np.ndarray))
    assert cur_bboxes.ndim == 2
    assert cur_bboxes.shape[1] == 4


def get_image_hw(image: Union[np.ndarray, torch.Tensor]) -> SizeHW:
    if isinstance(image, np.ndarray):
        image_hw = (image.shape[0], image.shape[1])
    elif isinstance(image, torch.Tensor):
        image_hw = (image.shape[-2], image.shape[-1])
    else:
        raise TypeError("Only array and tensor types are available")
    return image_hw


def get_karesize_param(src_hw: SizeHW, new_hw: SizeHW) -> Tuple[float, SizeHW, List[int]]:
    scale = min(new_hw[0] / src_hw[0], new_hw[1] / src_hw[1])
    resized_hw = (int(scale * src_hw[0]), int(scale * src_hw[1]))

    pad_left = (new_hw[1] - resized_hw[1]) // 2
    pad_right = new_hw[1] - resized_hw[1] - pad_left

    pad_top = (new_hw[0] - resized_hw[0]) // 2
    pad_bottom = new_hw[0] - resized_hw[0] - pad_top

    return scale, resized_hw, [pad_left, pad_right, pad_top, pad_bottom]


def resize_tensor(tensor: torch.Tensor, new_size_hw: SizeHW) -> torch.Tensor:
    assert isinstance(tensor, torch.Tensor)
    ndim = tensor.ndim
    assert 2 <= ndim <= 4

    dtype = tensor.dtype

    # Tensor must have BxCxHxW shape to interpolate its in spatial dimension
    for _ in range(4 - ndim):
        tensor = tensor.unsqueeze(0)

    min_value = torch.min(tensor).item()
    max_value = torch.max(tensor).item()

    tensor_resized = torch.nn.functional.interpolate(
        tensor.float(), (new_size_hw[0], new_size_hw[1]), align_corners=True, mode="bicubic"
    ).clamp(min_value, max_value)
    for _ in range(4 - ndim):
        tensor_resized = tensor_resized.squeeze(0)

    if dtype == torch.bool:
        if tensor.requires_grad:
            print(
                "Resizing for torch.bool is not implemented in pytorch. " "Gradients will be discarded due to resizing",
                flush=True,
            )
        thr = (max_value - min_value) / 2
        tensor_resized = tensor_resized >= thr

    tensor_resized = tensor_resized.to(dtype=dtype)

    return tensor_resized


def resize_ndarray(array: np.ndarray, new_size_hw: SizeHW) -> np.ndarray:
    assert isinstance(array, np.ndarray)
    ndim = array.ndim
    assert 2 <= ndim <= 3

    dtype = array.dtype

    if dtype == bool:
        array = array.astype(np.uint8)

    array_resized = cv2.resize(array, (new_size_hw[1], new_size_hw[0]), interpolation=cv2.INTER_CUBIC).astype(dtype)

    if array_resized.ndim < ndim:
        array_resized = array_resized[..., np.newaxis]

    return array_resized


__all__ = [
    "karesize_image",
    "inverse_karesize_image",
    "karesize_bboxes",
    "inverse_karesize_bboxes",
    "get_image_hw",
    "get_karesize_param",
    "resize_tensor",
    "resize_ndarray",
]



import warnings
from argparse import Namespace
from pathlib import Path
from typing import Any, Dict, List, Mapping, Optional, Union

import matplotlib.pyplot as plt
import numpy as np
from lightning_fabric.utilities.logger import _flatten_dict
from lightning_fabric.utilities.rank_zero import rank_zero_only
from pytorch_lightning.loggers import (
    Logger,
    MLFlowLogger,
    NeptuneLogger,
    TensorBoardLogger,
    WandbLogger,
)

from oml.const import OML_PATH, TCfg
from oml.interfaces.loggers import IPipelineLogger
from oml.registry.transforms import save_transforms_as_files
from oml.utils.images.images import figure_to_nparray
from oml.utils.misc import dictconfig_to_dict, flatten_dict


def prepare_config_to_logging(cfg: TCfg, sep: str = "|") -> Dict[str, Any]:
    cwd = Path.cwd().name
    flattened_dict = flatten_dict({**dictconfig_to_dict(cfg), **{"dir": cwd}}, sep=sep)
    return flattened_dict


def prepare_tags(cfg: TCfg) -> List[str]:
    cwd = Path.cwd().name
    tags = list(cfg.get("tags", [])) + [cfg.get("postfix", "")] + [cwd]
    tags = list(filter(lambda x: len(x) > 0, tags))
    return tags


class ClearMLLogger(Logger):
    def __init__(self, **kwargs: Any):
        try:
            from clearml import Task
        except ImportError as e:
            raise ModuleNotFoundError(
                "This contrib module requires clearml to be installed. "
                "You may install clearml using: \n pip install clearml \n"
            ) from e

        experiment_kwargs = {
            k: v for k, v in kwargs.items() if k not in ("project_name", "task_name", "task_type", "offline_mode")
        }

        if kwargs.get("offline_mode", True):
            Task.set_offline(offline_mode=True)
            warnings.warn("ClearMLSaver: running in offline mode")

        # Try to retrieve current the ClearML Task before trying to create a new one
        self.task = Task.current_task()
        if self.task is None:
            self.task = Task.init(
                project_name=kwargs.get("project_name"),
                task_name=kwargs.get("task_name"),
                task_type=kwargs.get("task_type", Task.TaskTypes.training),
                **experiment_kwargs,
            )

        self.logger = self.task.get_logger()

    @property
    def name(self) -> str:
        return "ClearMLLogger"

    @property
    def version(self) -> Union[int, str]:
        return self.task.id

    @rank_zero_only
    def finalize(self, status: str) -> None:
        self.logger.flush()

    @rank_zero_only
    def log_hyperparams(self, params: Optional[Union[Dict[str, Any], Namespace]]) -> None:
        if isinstance(params, Namespace):
            params = vars(params)

        if params is None:
            params = {}
        params = _flatten_dict(params)

        self.task.connect(params)

    @rank_zero_only
    def log_metrics(self, metrics: Mapping[str, float], step: Optional[int] = None) -> None:
        assert rank_zero_only.rank == 0, "experiment tried to log from global_rank != 0"  # type: ignore
        for k, v in metrics.items():
            self.logger.report_scalar(title=k, series=k, iteration=step, value=v)


class NeptunePipelineLogger(NeptuneLogger, IPipelineLogger):
    def log_pipeline_info(self, cfg: TCfg) -> None:
        warnings.warn(
            "Unfortunately, in the case of using Neptune, you may experience that long experiments are "
            "stacked and not responding. It's not an issue on OML's side, so, we cannot fix it."
        )
        self.log_hyperparams(prepare_config_to_logging(cfg))

        tags = prepare_tags(cfg)
        self.run["sys/tags"].add(tags)

        # log transforms as files
        for key, transforms_file in save_transforms_as_files(cfg):
            self.run[key].upload(transforms_file)

        # log source code
        source_files = list(map(lambda x: str(x), OML_PATH.glob("**/*.py"))) + list(
            map(lambda x: str(x), OML_PATH.glob("**/*.yaml"))
        )
        self.run["code"].upload_files(source_files)

        # log dataframe
        self.run["dataset"].upload(str(Path(cfg["dataset_root"]) / cfg["dataframe_name"]))

    def log_figure(self, fig: plt.Figure, title: str, idx: int) -> None:
        from neptune.types import File  # this is the optional dependency

        self.experiment[title].log(File.as_image(fig))


class WandBPipelineLogger(WandbLogger, IPipelineLogger):
    def log_pipeline_info(self, cfg: TCfg) -> None:
        # this is the optional dependency
        import wandb

        self.log_hyperparams(prepare_config_to_logging(cfg))

        tags = prepare_tags(cfg)
        self.experiment.tags = tags

        # log transforms as files
        keys_files = save_transforms_as_files(cfg)
        if keys_files:
            transforms = wandb.Artifact("transforms", type="transforms")
            for _, transforms_file in keys_files:
                transforms.add_file(transforms_file)
            self.experiment.log_artifact(transforms)

        # log source code
        code = wandb.Artifact("source_code", type="code")
        code.add_dir(OML_PATH, name="oml")
        self.experiment.log_artifact(code)

        # log dataset
        dataset = wandb.Artifact("dataset", type="dataset")
        dataset.add_file(str(Path(cfg["dataset_root"]) / cfg["dataframe_name"]))
        self.experiment.log_artifact(dataset)

    def log_figure(self, fig: plt.Figure, title: str, idx: int) -> None:
        fig_img = figure_to_nparray(fig)
        self.log_image(images=[fig_img], key=title)


class TensorBoardPipelineLogger(TensorBoardLogger, IPipelineLogger):
    def log_pipeline_info(self, cfg: TCfg) -> None:
        pass

    def log_figure(self, fig: plt.Figure, title: str, idx: int) -> None:
        fig_img = figure_to_nparray(fig)
        self.experiment.add_image(title, np.transpose(fig_img, (2, 0, 1)), idx)


class MLFlowPipelineLogger(MLFlowLogger, IPipelineLogger):
    def log_pipeline_info(self, cfg: TCfg) -> None:
        # log config
        self.log_hyperparams(prepare_config_to_logging(cfg, sep="/"))

        # log tags
        for tag in prepare_tags(cfg):
            self.experiment.set_tag(run_id=self.run_id, key=tag, value="tag")

        # log transforms as files
        names_files = save_transforms_as_files(cfg)
        if names_files:
            for name, transforms_file in names_files:
                self.experiment.log_artifact(run_id=self.run_id, local_path=transforms_file, artifact_path=name)

        # log code
        self.experiment.log_artifacts(run_id=self.run_id, local_dir=OML_PATH, artifact_path="code")

        # log dataframe
        self.experiment.log_artifact(
            run_id=self.run_id,
            local_path=str(Path(cfg["dataset_root"]) / cfg["dataframe_name"]),
            artifact_path="dataset",
        )

    def log_figure(self, fig: plt.Figure, title: str, idx: int) -> None:
        self.experiment.log_figure(figure=fig, artifact_file=f"{title}.png", run_id=self.run_id)


class ClearMLPipelineLogger(ClearMLLogger, IPipelineLogger):
    def log_pipeline_info(self, cfg: TCfg) -> None:
        # log config
        self.log_hyperparams(prepare_config_to_logging(cfg))

        # log tags
        self.task.add_tags(prepare_tags(cfg))

        # log transforms as files
        names_files = save_transforms_as_files(cfg)
        if names_files:
            for name, transforms_file in names_files:
                self.task.upload_artifact(name=name, artifact_object=transforms_file)

        # log code
        self.task.upload_artifact(name="code", artifact_object=OML_PATH)

        # log dataframe
        self.task.upload_artifact(
            name="dataset",
            artifact_object=str(Path(cfg["dataset_root"]) / cfg["dataframe_name"]),
        )

    def log_figure(self, fig: plt.Figure, title: str, idx: int) -> None:
        self.logger.report_matplotlib_figure(
            title=title,
            series="",
            figure=fig,
            iteration=idx,
            report_image=True,
        )


__all__ = [
    "IPipelineLogger",
    "TensorBoardPipelineLogger",
    "WandBPipelineLogger",
    "NeptunePipelineLogger",
    "MLFlowPipelineLogger",
    "ClearMLPipelineLogger",
]



import torchvision.transforms as t
from torchvision.transforms import Compose, InterpolationMode, Normalize, ToTensor

from oml.const import MEAN, STD, TNormParam


def get_augs_torch(im_size: int, mean: TNormParam = MEAN, std: TNormParam = STD) -> t.Compose:
    augs = t.Compose(
        [
            t.Resize(size=(im_size, im_size), antialias=True),
            t.RandomHorizontalFlip(),
            t.ColorJitter(),
            t.ToTensor(),
            t.Normalize(mean=mean, std=std),
        ]
    )
    return augs


def get_augs_hypvit(
    im_size: int = 224, min_scale: float = 0.2, mean: TNormParam = MEAN, std: TNormParam = STD
) -> t.Compose:
    augs = t.Compose(
        [
            t.RandomResizedCrop(im_size, scale=(min_scale, 1.0), interpolation=InterpolationMode.BICUBIC),
            t.RandomHorizontalFlip(),
            t.ToTensor(),
            t.Normalize(mean=mean, std=std),
        ]
    )
    return augs


def get_normalisation_resize_hypvit(
    im_size: int = 224, crop_size: int = 224, mean: TNormParam = MEAN, std: TNormParam = STD
) -> t.Compose:
    transforms = t.Compose(
        [
            t.Resize(im_size, interpolation=InterpolationMode.BICUBIC),
            t.CenterCrop(crop_size),
            t.ToTensor(),
            t.Normalize(mean=mean, std=std),
        ]
    )
    return transforms


def get_normalisation_torch(mean: TNormParam = MEAN, std: TNormParam = STD) -> Compose:
    return Compose([ToTensor(), Normalize(mean=mean, std=std)])


def get_normalisation_resize_torch(im_size: int, mean: TNormParam = MEAN, std: TNormParam = STD) -> Compose:
    return Compose([t.Resize(size=(im_size, im_size), antialias=True), ToTensor(), Normalize(mean=mean, std=std)])


__all__ = [
    "get_augs_torch",
    "get_normalisation_torch",
    "get_normalisation_resize_torch",
    "get_augs_hypvit",
    "get_normalisation_resize_hypvit",
]



from abc import ABC, abstractmethod
from typing import Any, Collection, Dict, List, Tuple, Union

import matplotlib.pyplot as plt
from torch import LongTensor

from oml.const import OVERALL_CATEGORIES_KEY

TIndices = Union[LongTensor, List[int]]


class IBasicMetric(ABC):
    """
    This is a base interface for the objects that calculate metrics.

    """

    metric_name = "BASE"
    overall_categories_key = OVERALL_CATEGORIES_KEY

    @abstractmethod
    def setup(self, *args: Any, **kwargs: Any) -> Any:
        """
        Method for preparing metrics to work: memory allocation, placeholder preparation, etc.
        Has to be called before the first call of ``self.update_data()``.

        """
        raise NotImplementedError()

    @abstractmethod
    def update_data(self, data: Dict[str, Any], indices: TIndices) -> None:
        """
        Method for passing data to calculate the metrics later on.

        """
        raise NotImplementedError()

    @abstractmethod
    def compute_metrics(self, *args: Any, **kwargs: Any) -> Any:
        """
        The output must be in the following format:

        .. code-block:: python

            {
                "self.overall_categories_key": {"metric1": ..., "metric2": ...},
                "category1": {"metric1": ..., "metric2": ...},
                "category2": {"metric1": ..., "metric2": ...}
            }

        Where ``category1`` and ``category2`` are optional.
        """

        raise NotImplementedError()


class IMetricVisualisable(IBasicMetric):
    """
    This is an interface for all metrics which can visualize themselves.

    """

    @abstractmethod
    def visualize(self) -> Tuple[Collection[plt.Figure], Collection[str]]:
        """
        Method which returns results of visualization and titles for logging.

        """
        raise NotImplementedError()

    @abstractmethod
    def ready_to_visualize(self) -> bool:
        """
        Method which checks if visualization can be done.

        """
        raise NotImplementedError()


__all__ = ["IBasicMetric", "IMetricVisualisable"]



import os
from pathlib import Path

from oml.const import PROJECT_ROOT


def test_readme_was_built_correctly() -> None:
    readme_tmp_file = "README_tmp"
    os.chdir(PROJECT_ROOT)
    os.system(f"make build_readme README_FILE={readme_tmp_file}.md")

    with open("README.md", "r") as f1:
        readme = f1.read()

    with open(f"{readme_tmp_file}.md", "r") as f2:
        readme_tmp = f2.read()

    assert readme == readme_tmp, (
        "Automatically generated readme and yours are not the same. The reason is that"
        "you tried to change the README.md inplace, instead of changing its source components"
        "in docs/readme folder."
    )

    Path(f"{readme_tmp_file}.md").unlink()



from pathlib import Path
from typing import Tuple, Union

import gdown
import pandas as pd

from oml.const import (
    CATEGORIES_COLUMN,
    IS_GALLERY_COLUMN,
    IS_QUERY_COLUMN,
    LABELS_COLUMN,
    MOCK_DATASET_MD5,
    MOCK_DATASET_PATH,
    MOCK_DATASET_URL_GDRIVE,
    SPLIT_COLUMN,
    TEXTS_COLUMN,
)
from oml.utils.io import check_exists_and_validate_md5
from oml.utils.remote_storage import download_folder_from_remote_storage


def get_mock_images_dataset(
    dataset_root: Union[str, Path] = MOCK_DATASET_PATH,
    check_md5: bool = True,
    df_name: str = "df.csv",
    global_paths: bool = False,
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """
    Function to download mock dataset which is already prepared in the required format.

    Args:
        dataset_root: Path to save the dataset
        check_md5: Set ``True`` to check md5sum
        df_name: Name of csv file for which output DataFrames will be returned
        global_paths: Set ``True`` to cancat paths and ``dataset_root``

    Returns: Dataframes for the training and validation stages

    """
    dataset_root = Path(dataset_root)

    if not check_exists_and_validate_md5(dataset_root, MOCK_DATASET_MD5 if check_md5 else None):
        try:
            print("Downloading from oml.daloroserver.com")
            download_folder_from_remote_storage(remote_folder=MOCK_DATASET_PATH.name, local_folder=str(dataset_root))
            assert check_exists_and_validate_md5(dataset_root, MOCK_DATASET_MD5 if check_md5 else None)
        except Exception:
            print("We could not download from oml.daloroserver.com, let's try Google Drive.")
            gdown.download_folder(url=MOCK_DATASET_URL_GDRIVE, output=str(dataset_root))
    else:
        print(f"Mock dataset has been downloaded already to {dataset_root}")

    if not check_exists_and_validate_md5(dataset_root, MOCK_DATASET_MD5 if check_md5 else None):
        raise Exception("Downloaded mock dataset is invalid.")

    df = pd.read_csv(Path(dataset_root) / df_name)

    if global_paths:
        df["path"] = df["path"].apply(lambda x: str(Path(dataset_root) / x))

    df_train = df[df["split"] == "train"].reset_index(drop=True)
    df_val = df[df["split"] == "validation"].reset_index(drop=True)

    df_val["is_query"] = df_val["is_query"].astype(bool)
    df_val["is_gallery"] = df_val["is_gallery"].astype(bool)

    return df_train, df_val


def download_mock_dataset(
    dataset_root: Union[str, Path] = MOCK_DATASET_PATH,
    check_md5: bool = True,
    df_name: str = "df.csv",
    global_paths: bool = False,
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    # for back compatibility

    return get_mock_images_dataset(
        dataset_root=dataset_root, check_md5=check_md5, df_name=df_name, global_paths=global_paths
    )


def get_mock_texts_dataset() -> Tuple[pd.DataFrame, pd.DataFrame]:
    """
    Mock texts dataset useful for prototyping pipelines and understanding dataset structure.

    """

    bed = [
        "Luxura King Bed: Plush headboard and durable frame for ultimate comfort.",
        "EcoSleep Twin Bed: Sustainable materials for eco-conscious consumers.",
        "DreamRest Queen Bed: Ergonomic design for restful sleep.",
        "ClassicWood Full Bed: Sturdy wood construction with a sleek finish.",
    ]

    table = [
        "UrbanChic Dining Table: Glass top with metal legs seats six.",
        "RusticFarmhouse Coffee Table: Reclaimed wood for rustic charm.",
        "Minimalist Work Desk: Sleek, modern workspace.",
        "ArtisanCraft End Table: Handcrafted accent piece.",
        "Vintage Bistro Table: Classic design for cozy corners.",
        "Compact Folding Table: Versatile and easy to store.",
    ]

    tv = [
        'UltraHD Smart TV 55": Stunning visuals and smart features.',
        'Compact LED TV 32": Crisp picture in a space-saving design.',
        'Curved 4K TV 65": Panoramic viewing with brilliant colors.',
        'BudgetFriendly LCD TV 40": Clear picture and essential features.',
    ]

    potato = [
        "Yukon Gold Potato: Creamy and golden, perfect for roasting and mashing.",
        "Russet Potato: Classic and versatile, ideal for baking and frying.",
        "Red Potato: Smooth and firm, great for boiling and salads.",
        "Fingerling Potato: Rich and nutty, perfect for roasting or grilling.",
        "Purple Potato: Vibrant and sweet, adds color to any dish.",
    ]

    chair = [
        "ErgoComfort Office Chair: Maximum comfort and support.",
        "ClassicWood Dining Chair: Timeless design with sturdy construction.",
        "RelaxLounge Recliner: Adjustable settings and plush cushioning.",
        "ModernAccent Chair: Contemporary style with vibrant color options.",
    ]

    phone = [
        "Galaxy X9: Sleek design with powerful performance.",
        "iPhone 12: Cutting-edge technology with a stylish look.",
        "Pixel 5: Pure Android experience with excellent camera.",
        "OnePlus 8T: High performance at a competitive price.",
        "Moto G Power: Long battery life for extended use.",
        "Sony Xperia 5: Superior camera and display quality.",
    ]

    audio = [
        "NoiseCancelling Headphones Pro: Superior sound and noise cancellation.",
        "Wireless Earbuds Sport: Great sound with a secure fit.",
        "StudioOverEar Headphones: Exceptional audio clarity.",
        "BudgetFriendly Wired Headphones: Good sound at an affordable price.",
        "Gaming Headset Pro: Surround sound for gaming.",
        "TravelNoiseCancelling Earbuds: Compact with excellent noise cancellation.",
    ]
    texts = [*bed, *table, *tv, *potato, *chair, *phone, *audio]

    labels = (
        [0] * len(bed)
        + [1] * len(table)
        + [2] * len(tv)
        + [3] * len(potato)
        + [4] * len(chair)
        + [5] * len(phone)
        + [6] * len(audio)
    )

    categories = (
        ["furniture"] * len(bed)
        + ["furniture"] * len(table)
        + ["electronic"] * len(tv)
        + ["food"] * len(potato)
        + ["furniture"] * len(chair)
        + ["electronic"] * len(phone)
        + ["electronic"] * len(audio)
    )

    n_train = len(bed) + len(table) + len(tv) + len(potato)
    n_val = len(chair) + len(phone) + len(audio)

    split = ["train"] * n_train + ["validation"] * n_val
    is_query = [None] * n_train + [True] * n_val  # type: ignore

    is_gallery = is_query

    data = {
        TEXTS_COLUMN: texts,
        LABELS_COLUMN: labels,
        CATEGORIES_COLUMN: categories,
        SPLIT_COLUMN: split,
        IS_QUERY_COLUMN: is_query,
        IS_GALLERY_COLUMN: is_gallery,
    }

    df = pd.DataFrame(data)

    df_train = df[df["split"] == "train"].reset_index(drop=True)
    df_val = df[df["split"] == "validation"].reset_index(drop=True)

    return df_train, df_val


__all__ = ["get_mock_images_dataset", "get_mock_texts_dataset"]
